"""
Helper function
"""
from numpy.random import randint

def swap_single_row(x):
    new_x = x
    idx = randint(0, len(x), size=2)
    new_idx = idx[::-1]
    new_x[new_idx] = x[idx]
    return new_x


#!/usr/bin/env python
# -*- coding: latin-1 -*-

#
# fs_shapeDNA
#
# script to compute ShapeDNA of FreeSurfer structures
#
# Original Author: Martin Reuter
# Date: Dec-16-2014
#

import warnings
import os
import sys
import shlex
import optparse
import subprocess
import tempfile
import uuid
import errno

warnings.filterwarnings('ignore', '.*negative int.*')
os.environ['OMP_NUM_THREADS'] = '1' # setenv OMP_NUM_THREADS 1

def my_print(message):
    """
    print message, then flush stdout
    """
    print(message)
    sys.stdout.flush()

HELPTEXT = """

SUMMARY

Computes surface (and volume) models of FreeSurfer's subcortical and
cortical structures and computes a spectral shape descriptor
(ShapeDNA [1]).

Input can be one of the following:
--asegid  : an aseg label id to construct a surface of that ROI
--surf    : a surface, e.g. lh.white
--label   : a label file to cut out a surface patch (also pass --surf)
--aparcid : an aparc id to cut out that cortical ROI (also pass --surf)

If "meshfix" and "gmsh" are in the path, and shapeDNA-tetra is available
in the $SHAPEDNA_HOME directory, it can also create and process
tetrahedral meshes from closed surfaces, such as lh.white ( --dotet ).

Note both FreeSurfer needs to be sourced and the environment variable
$SHAPEDNA_HOME needs to be set to point to the shapeDNA executables
and key file.

REFERENCES
==========

Always cite [1] as it describes the method. If you use topological
features of eigenfunctions, also cite [2]. [3] compares different
discretizations of Laplace-Beltrami operators and shows that the
used FEM appraoch performs best. If you do statistical shape analysis
you may also want to cite [4] as it discusses medical applications.

[1] M. Reuter, F.-E. Wolter and N. Peinecke.
Laplace-Beltrami spectra as "Shape-DNA" of surfaces and solids.
Computer-Aided Design 38 (4), pp.342-366, 2006.
http://dx.doi.org/10.1016/j.cad.2005.10.011

[2] M. Reuter. Hierarchical Shape Segmentation and Registration
via Topological Features of Laplace-Beltrami Eigenfunctions.
International Journal of Computer Vision, 2009.
http://dx.doi.org/10.1007/s11263-009-0278-1

[3] M. Reuter, S. Biasotti, D. Giorgi, G. Patane, M. Spagnuolo.
Discrete Laplace-Beltrami operators for shape analysis and
segmentation. Computers & Graphics 33 (3), pp.381-390, 2009.
http://dx.doi.org/10.1016/j.cag.2009.03.005

[4] M. Reuter, F.-E. Wolter, M. Shenton, M. Niethammer.
Laplace-Beltrami Eigenvalues and Topological Features of
Eigenfunctions for Statistical Shape Analysis.
Computer-Aided Design 41 (10), pp.739-755, 2009.
http://dx.doi.org/10.1016/j.cad.2009.02.007
"""

TMPTXT="""
REQUIRED ARGUMENTS

--sid <name>      Subject ID

One of the following:

--asegid <int>    Segmentation ID of structure in aseg.mgz (e.g. 17 is
                  Left-Hippocampus), for ID's check <sid>/stats/aseg.stats
                  or $FREESURFER_HOME/FreeSurferColorLUT.txt.

--surf <name>     lh.pial, rh.pial, lh.white, rh.white etc. to select a 
                  surface from the <sid>/surfs directory.


OPTIONAL ARGUMENTS

--sdir <name>     Subjects directory (or set via environment $SUBJECTS_DIR)

--outdir <name>   Output directory (default: <sdir>/<sid>/brainprint/ )

--outsurf <name>  Name for surface output (with --asegid)
                  (default: aseg.<asegid>.surf )

--outev <name>    Name for eigenvalue output
                  (default: <(out)surf>.<shapedna_parametrs>.ev )

ShapeDNA parameters (see shapeDNA-tria --help for details):

--num <int>       Number of eigenvalues/vectors to compute (default: 50)

--degree <int>    FEM degree (default 1)

--bcond <int>     Boundary condition (0=Dirichlet, 1=Neumann default)

--evec            Additionally compute eigenvectors

--ignorelq        Ignore low quality in input mesh

--sparam "<str>"  Additional parameters for shapeDNA-tria
"""

def split_callback(option, opt, value, parser):
  setattr(parser.values, option.dest, value.split(','))

def options_parse():
    """
    Command Line Options Parser:
    initiate the option parser and return the parsed object
    """
    parser = optparse.OptionParser(version='$Id: fs_shapeDNA,v 1.21 2013/05/17 15:14:08 mreuter Exp $', usage=HELPTEXT)
    
    # help text
    h_sid       = '(REQUIRED) subject ID (FS processed directory inside the subjects directory)'
    h_sdir      = 'FS subjects directory (or set environment $SUBJECTS_DIR)'
    
    h_asegid    = 'segmentation ID of structure in aseg.mgz (e.g. 17 is Left-Hippocampus), for ID\'s check <sid>/stats/aseg.stats or $FREESURFER_HOME/FreeSurferColorLUT.txt'
    h_surf      = 'surface name, e.g. lh.pial, rh.pial, lh.white, rh.white etc. to select a surface from the <sid>/surfs directory'
    h_aparcid   = 'segmentation ID of structure in aparc (e.g. 24 is precentral), requires --surf (including the hemi prefix), for ID\'s check $FREESURFER_HOME/average/colortable_desikan_killiany.txt'
    h_label     = 'full path to label file, to create surface patch, requires --surf (including the hemi prefix)'
    h_source    = 'specify source subject with --label (to map label e.g. from fsaverage space)'

    h_dotet     = 'construct a tetrahedra volume mesh and compute spectra of solid'
    h_fixiter   = 'iterations of meshfix (default=4), only with --dotet'
    
    h_outdir    = 'Output directory (default: <sdir>/<sid>/brainprint/ )'
    h_outsurf   = 'Full path for surface output in VTK format (with --asegid default: <outdir>/aseg.<asegid>.vtk )'
    h_outtet    = 'Full path for tet output (with --dotet) (default: <outdir>/<(out)surf>.msh )'
    h_outev     = 'Full path for eigenvalue output (default: <outdir>/<(out)surf or outtet>.ev )'
    
    h_num       = 'number of eigenvalues/vectors to compute (default: 50)'
    h_degree    = 'degree for FEM computation (1=linear default, 3=cubic)'
    h_bcond     = 'boundary condition (1=Neumann default, 0=Dirichlet )'
    h_evec      = 'bool to switch on eigenvector computation'
    h_ignorelq  = 'ignore low quality in input mesh'
    h_refmin    = 'mesh refinement so that DOF is at least <int>'
    h_tsmooth   = 'tangential smoothing iterations (after refinement)'
    h_gsmooth   = 'geometry smoothing iterations (after refinement)'
    h_param2d   = 'additional parameters for shapeDNA-tria'
   
    # Add options 

    # Sepcify inputs
    parser.add_option('--sdir', dest='sdir', help=h_sdir)

    group = optparse.OptionGroup(parser, "Required Options", "Specify --sid and select one of the other options")
    group.add_option('--sid',     dest='sid',      help=h_sid)
    group.add_option('--asegid',  dest='asegid' ,  help=h_asegid,  type='string', action='callback', callback=split_callback)
    group.add_option('--aparcid', dest='aparcid' , help=h_aparcid, type='string', action='callback', callback=split_callback)
    group.add_option('--surf' ,   dest='surf' ,    help=h_surf)
    group.add_option('--label' ,  dest='label' ,   help=h_label)
    parser.add_option_group(group)


    group = optparse.OptionGroup(parser, "Additional Flags", )
    group.add_option('--source' , dest='source',   help=h_source)
    group.add_option('--dotet' ,  dest='dotet',    help=h_dotet,   action='store_true', default=False)
    group.add_option('--fixiter', dest='fixiter',  help=h_fixiter, default=4, type='int')
    parser.add_option_group(group)

    #output switches
    group = optparse.OptionGroup(parser, "Output Parameters" )
    group.add_option('--outdir',  dest='outdir',   help=h_outdir)
    group.add_option('--outsurf', dest='outsurf',  help=h_outsurf)
    group.add_option('--outtet',  dest='outtet',   help=h_outtet)
    group.add_option('--outev',   dest='outev',    help=h_outev)
    parser.add_option_group(group)

    #shapedna switches
    group = optparse.OptionGroup(parser, "ShapeDNA Parameters","See shapeDNA-tria --help for details")
    group.add_option('--num' ,     dest='num',      help=h_num,      default=50, type='int')
    group.add_option('--degree' ,  dest='degree',   help=h_degree,   default=1,  type='int')
    group.add_option('--bcond' ,   dest='bcond',    help=h_bcond,    default=1,  type='int')
    group.add_option('--evec' ,    dest='evec',     help=h_evec,     default=False, action='store_true' )
    group.add_option('--ignorelq', dest='ignorelq', help=h_ignorelq, default=False, action='store_true')
    group.add_option('--refmin',   dest='refmin',   help=h_refmin,   default=0,  type='int')
    group.add_option('--tsmooth',  dest='tsmooth',  help=h_tsmooth,  default=0,  type='int')   
    group.add_option('--gsmooth',  dest='gsmooth',  help=h_gsmooth,  default=0,  type='int')   
    group.add_option('--param2d',  dest='param2d',  help=h_param2d)
    parser.add_option_group(group)
    
                      
    (options, args) = parser.parse_args()
    #if len(args) == 0:
    #    parser.print_help()
    #    my_print('\n')
    #    sys.exit(1)
                
    # WITHOUT FREESURFER DO NOTHING
    fshome = os.getenv('FREESURFER_HOME')
    if fshome is None:
        parser.print_help()
        my_print('\nERROR: Environment variable FREESURFER_HOME not set. \n')
        my_print('       Make sure to source your FreeSurfer installation.\n')
        sys.exit(1)

    sdnahome = os.getenv('SHAPEDNA_HOME')
    if sdnahome is None:
        parser.print_help()
        my_print('\nERROR: Environment variable SHAPEDNA_HOME not set.\n')
        my_print('       Set SHAPEDNA_HOME to point to the shapeDNA-tria installation.\n')
        sys.exit(1)

    sdna = os.path.join(sdnahome,"shapeDNA-tria")
    if not os.path.exists(sdna) and not options.dotet:
        parser.print_help()
        my_print('\nERROR: Cannot find shapeDNA-tria\n')
        my_print('       Make sure that binary is at $SHAPEDNA_HOME \n')
        sys.exit(1)

    if which("shapeDNA-tetra") is None and options.dotet:
        parser.print_help()
        my_print('\nERROR: Cannot find shapeDNA-tetra\n')
        my_print('       Make sure that binary is at $SHAPEDNA_HOME \n')
        sys.exit(1)

    if options.sdir is None:
        options.sdir = os.getenv('SUBJECTS_DIR')

    if options.sdir is None:
        parser.print_help()
        my_print('\nERROR: specify subjects directory via --sdir or $SUBJECTS_DIR\n')
        sys.exit(1)
        
    if options.sid is None:
        parser.print_help()
        my_print('\nERROR: Specify --sid\n')
        sys.exit(1)
            
    subjdir = os.path.join(options.sdir,options.sid)
    if not os.path.exists(subjdir):
        my_print('ERROR: cannot find sid in subjects directory\n')
        sys.exit(1)
    
    if options.label is not None and options.surf is None:
        parser.print_help()
        my_print('\nERROR: Specify --surf with --label\n')
        sys.exit(1)  
    if options.aparcid is not None and options.surf is None:
        parser.print_help()
        my_print('\nERROR: Specify --surf with --aparc\n')
        sys.exit(1)  
    # input needs to be either a surf or aseg label(s)
    if options.asegid is None and options.surf is None:
        parser.print_help()
        my_print('\nERROR: Specify either --asegid or --surf\n')
        sys.exit(1)
    # and it cannot be both
    if options.asegid is not None and options.surf is not None:
        parser.print_help()
        my_print('\nERROR: Specify either --asegid or --surf (not both)\n')
        sys.exit(1)  
    
    # set default output dir (maybe this should be ./ ??)
    if options.outdir is None:
        options.outdir = os.path.join(subjdir,'brainprint')
    try:
        os.mkdir(options.outdir)
    except OSError as e:
        if e.errno != os.errno.EEXIST:
            raise e
        pass

    # check if we have write access to output dir
    try:
        testfile = tempfile.TemporaryFile(dir = options.outdir)
        testfile.close()
    except OSError as e:
        if e.errno != errno.EACCES:  # 13
            e.filename = options.outdir
            raise
        my_print('\nERROR: '+options.outdir+' not writeable (check access)!\n')
        sys.exit(1)
    
    # initialize outsurf
    if options.outsurf is None:
        # for aseg stuff, we need to create a surface (and we'll keep it around)
        if options.asegid is not None:
            astring  = '_'.join(options.asegid)
            #surfname = 'aseg.'+astring+'.surf'
            surfname = 'aseg.'+astring+'.vtk'
            options.outsurf = os.path.join(options.outdir,surfname)    
        elif options.label is not None:
            surfname = os.path.basename(options.surf)+'.'+os.path.basename(options.label)+'.vtk'
            options.outsurf = os.path.join(options.outdir,surfname)
        elif options.aparcid is not None:
            astring  = '_'.join(options.aparcid)
            surfname = os.path.basename(options.surf)+'.aparc.'+astring+'.vtk'
            options.outsurf = os.path.join(options.outdir,surfname)          
        else:
            # for surfaces, a refined/smoothed version could be written
            surfname = os.path.basename(options.surf)+'.vtk'
            options.outsurf = os.path.join(options.outdir,surfname)
    else:
        # make sure it is vtk ending
        if (os.path.splitext(options.outsurf)[1]).upper() != '.VTK':
            my_print('ERROR: outsurf needs vtk extension (VTK format)')
            sys.exit(1)
    
    # for 3d processing, initialize outtet:
    if options.dotet and options.outtet is None:
        surfname = os.path.basename(options.outsurf)
        options.outtet = os.path.join(options.outdir,surfname+'.msh')    
    
    # set source to sid if empty
    if options.source is None:
        options.source = options.sid
        
    # if label does not exist, search in subject label dir
    if options.label is not None and not os.path.isfile(options.label):
        ltemp = os.path.join(options.sdir,options.source,'label',options.label)
        if os.path.isfile(ltemp):
            options.label = ltemp
        else:
            parser.print_help()
            my_print('\nERROR: Specified --label not found\n')
            sys.exit(1)  
                
    # initialize outev 
    if options.outev is None:
        if options.dotet:
            options.outev = options.outtet+'.ev'
        else:
            options.outev = options.outsurf+'.ev'
        
    return options

def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
        if is_exe(os.path.join(os.getenv('SHAPEDNA_HOME'),program)):
            return os.path.join(os.getenv('SHAPEDNA_HOME'),program)
        if is_exe(os.path.join('.',program)):
            return os.path.join('.',program)

    return None

def run_cmd(cmd,err_msg):
    """
    execute the comand
    """
    clist = cmd.split()
    progname=which(clist[0])
    if (progname) is None:
        my_print('ERROR: '+ clist[0] +' not found in path!')
        sys.exit(1)
    clist[0]=progname
    cmd = ' '.join(clist)
    my_print('#@# Command: ' + cmd+'\n')

    args = shlex.split(cmd)
    try:
        subprocess.check_call(args)
    except subprocess.CalledProcessError as e:
        my_print('ERROR: '+err_msg)
        #sys.exit(1)
        raise
    my_print('\n')

# Gets global path to surface input (if it is a FS surf)
def get_surf_surf(sdir,sid,surf):
    return os.path.join(sdir,sid,'surf',surf)


# creates tria surface from label (and specified surface)
# maps the label first if source is different from sid
def get_label_surf(sdir,sid,label,surf,source,outsurf):
    subjdir  = os.path.join(sdir,sid)
    outdir   = os.path.dirname( outsurf )
    stlsurf  = os.path.join(outdir,'label'+str(uuid.uuid4())+'.stl')

    # get hemi from surf
    splitsurf = surf.split(".",1)
    hemi = splitsurf[0]
    surf = splitsurf[1]
    # map label if necessary
    mappedlabel = label
    if (source != sid):
        mappedlabel = os.path.join(outdir,os.path.basename(label)+'.'+str(uuid.uuid4())+'.label')
        cmd = 'mri_label2label --sd '+sdir+' --srclabel '+label+' --srcsubject '+ source+' --trgsubject '+sid+' --trglabel '+mappedlabel+' --hemi '+hemi+' --regmethod surface'
        run_cmd(cmd,'mri_label2label failed?')     
    # make surface path (make sure output is stl, this currently needs fsdev, will probably be in 6.0)
    cmd = 'label2patch -writesurf -sdir '+sdir+' -surf '+ surf + ' ' +sid+' '+hemi+' '+mappedlabel+' '+stlsurf
    run_cmd(cmd,'label2patch failed?')     
    cmd = 'mris_convert '+stlsurf+' '+outsurf
    run_cmd(cmd,'mris_convert failed.')
    
    # cleanup map label if necessary
    if (source != sid):
        cmd ='rm '+mappedlabel
        run_cmd(cmd,'rm mapped label '+mappedlabel+' failed.')     
    cmd = 'rm '+stlsurf
    run_cmd(cmd,'rm stlsurf failed.')
    # return surf name
    return outsurf


# Creates a surface from the aseg and label info
# and writes it to the outdir
def get_aseg_surf(sdir,sid,asegid,outsurf):
    astring2 = ' '.join(asegid)
    subjdir  = os.path.join(sdir,sid)
    aseg     = os.path.join(subjdir,'mri','aseg.mgz')
    norm     = os.path.join(subjdir,'mri','norm.mgz')  
    outdir   = os.path.dirname( outsurf )    
    tmpname  = 'aseg.'+str(uuid.uuid4())
    segf     = os.path.join(outdir,tmpname+'.mgz')
    segsurf  = os.path.join(outdir,tmpname+'.surf')
    # binarize on selected labels (creates temp segf)
    ptinput = aseg
    ptlabel = str(asegid[0])
    #if len(asegid) > 1:
    # always binarize first, otherwise pretess may scale aseg if labels are larger than 255 (e.g. aseg+aparc, bug in mri_pretess?)
    cmd ='mri_binarize --i '+aseg+' --match '+astring2+' --o '+segf
    run_cmd(cmd,'mri_binarize failed.') 
    ptinput = segf
    ptlabel = '1'
    # if norm exist, fix label (pretess)
    if os.path.isfile(norm):
        cmd ='mri_pretess '+ptinput+' '+ptlabel+' '+norm+' '+segf
        run_cmd(cmd,'mri_pretess failed.') 
    else:
        if not os.path.isfile(segf):
            # cp segf if not exist yet
            # (it exists already if we combined labels above)
            cmd = 'cp '+ptinput+' '+segf
            run_cmd(cmd,'cp segmentation file failed.') 
    # runs marching cube to extract surface
    cmd ='mri_mc '+segf+' '+ptlabel+' '+segsurf
    run_cmd(cmd,'mri_mc failed?') 
    # convert to stl
    cmd ='mris_convert '+segsurf+' '+outsurf
    run_cmd(cmd,'mris_convert failed.')
    # cleanup temp files
    cmd ='rm '+segf
    run_cmd(cmd,'rm temp segf failed.') 
    cmd ='rm '+segsurf
    run_cmd(cmd,'rm temp segsurf failed.') 
    # return surf name
    return outsurf


# Creates a surface from the aparc and label number
# and writes it to the outdir
def get_aparc_surf(sdir,sid,surf,aparcid,outsurf):
    astring2 = ' '.join(aparcid)
    subjdir  = os.path.join(sdir,sid)
    outdir   = os.path.dirname( outsurf )    
    rndname = str(uuid.uuid4()) 
    # get hemi from surf
    hemi = surf.split(".",1)[0]
    # convert annotation id to label:
    alllabels = ''
    for aid in aparcid:
        # create label of this id
        outlabelpre = os.path.join(outdir,hemi+'.aparc.'+rndname)
        cmd = 'mri_annotation2label --sd '+sdir+' --subject '+sid+' --hemi '+hemi+' --label '+str(aid)+' --labelbase '+outlabelpre 
        run_cmd(cmd,'mri_annotation2label failed?') 
        alllabels = alllabels+'-i '+outlabelpre+"-%03d.label" % int(aid)+' ' 
    # merge labels (if more than 1)
    mergedlabel = outlabelpre+"-%03d.label" % int(aid)
    if len(aparcid) > 1:
        mergedlabel = os.path.join(outdir,hemi+'.aparc.all.'+rndname+'.label')
        cmd = 'mri_mergelabels '+alllabels+' -o '+mergedlabel
        run_cmd(cmd,'mri_mergelabels failed?') 
    # make to surface (call subfunction above)
    get_label_surf(sdir,sid,mergedlabel,surf,sid,outsurf)
    # cleanup
    if len(aparcid) > 1:
        cmd ='rm '+mergedlabel
        run_cmd(cmd,'rm '+mergedlabel+' failed?')
    for aid in aparcid:
        outlabelpre = os.path.join(outdir,hemi+'.aparc.'+rndname+"-%03d.label" % int(aid))
        cmd ='rm '+outlabelpre
        run_cmd(cmd,'rm '+outlabelpre+' failed?')
    # return surf name
    return outsurf

def get_tetmesh(surf,outtet,fixiter):
    surfbase  = os.path.basename(surf)
    outdir    = os.path.dirname( outtet )    
    surftemp_stl = os.path.join(outdir,surfbase+'.temp.stl')
    
    # massage surface mesh (rm handles, 60000 vertices, uniform)
    cmd='mris_convert '+surf+' '+surftemp_stl
    run_cmd(cmd,'mris_convert (to STLs) failed')
#    cmd='meshfix '+surftemp_stl+' -a 2.0 --remove-handles -q --stl -o '+surftemp_stl
#    run_cmd(cmd,'meshfix (remove-handles) failed, is it in your path?') 
    cmd='meshfix '+surftemp_stl+' -a 2.0 -u 5 --vertices 60000 -q --stl -o '+surftemp_stl
    run_cmd(cmd,'meshfix (downsample) failed?') 
    for num in range(0,fixiter):
        cmd='meshfix '+surftemp_stl+' -a 2.0 -u 1 -q --stl -o '+surftemp_stl
        run_cmd(cmd,'meshfix failed ('+str(num)+'a)?')      
        cmd='meshfix '+surftemp_stl+' -a 2.0 -q --stl -o '+surftemp_stl
        run_cmd(cmd,'meshfix failed ('+str(num)+'b)?') 
        
# replacing meshfix never worked:
#    sdnahome = os.getenv('SHAPEDNA_HOME')
#    triaio = os.path.join(sdnahome,"triaIO")  
#    cmd=triaio+' --infile '+surf+' --outfile '+surftemp_stl' --contractedges --remeshbk 1'
#    run_cmd(cmd,'triaIO remeshing failed?')
      
    # write gmsh geofile
    geofile  = os.path.join(outdir,'gmsh.'+str(uuid.uuid4())+'.geo')
    g = open(geofile, 'w')
    g.write("Mesh.Algorithm3D=4;\n")
    g.write("Mesh.Optimize=1;\n")
    g.write("Mesh.OptimizeNetgen=1;\n")
    g.write("Merge \"" + surfbase+'.temp.stl' + "\";\n")
    g.write("Surface Loop(1) = {1};\n")
    g.write("Volume(1) = {1};\n")
    g.write("Physical Volume(1) = {1};\n")
    g.close()
    # use gmsh to create tet mesh
    cmd = 'gmsh -3 -o '+outtet+' '+geofile
    run_cmd(cmd,'gmsh failed, is it in your path?') 
    
    # cleanup
    cmd ='rm '+geofile
    run_cmd(cmd,'rm temp geofile failed?') 
    cmd='rm '+surftemp_stl
    run_cmd(cmd,'rm temp stl surface failed?') 

    # return tetmesh filename
    return outtet

def run_shapeDNAtetra(tetmesh,outev,options):
    sdnahome = os.getenv('SHAPEDNA_HOME')
    sdna = os.path.join(sdnahome,"shapeDNA-tetra")
    # mesh refine degree num bcond ouput ev
    num = 50
    if options.num is not None:       
        num = options.num
    degree = 1
    if options.degree is not None:
        degree = options.degree
    dirichlet = ""
    if options.bcond == 0:
        dirichlet = " --dirichlet"
    cmd =sdna+' --mesh '+tetmesh+' --num '+str(options.num)+' --outfile '+outev+' --degree '+str(options.degree)+dirichlet
    if options.evec:
        cmd = cmd+' --evec'
    run_cmd(cmd,'shapeDNA-tetra failed?') 


def run_shapeDNAtria(surf,outev,outsurf,options):
 # options : num, degree, evec, ignorelq, refmin, tsmooth, gsmooth, param2d
    sdnahome = os.getenv('SHAPEDNA_HOME')
    sdna = os.path.join(sdnahome,"shapeDNA-tria") 
    num = 50
    if options.num is not None:       
        num = options.num
    degree = 1
    if options.degree is not None:
        degree = options.degree
    dirichlet = ""
    if options.bcond == 0:
        dirichlet = " --dirichlet"
    cmd =sdna+' --mesh '+surf+' --num '+str(options.num)+' --outfile '+outev+' --degree '+str(options.degree)+dirichlet
    if options.evec:
        cmd = cmd+' --evec'
    if options.ignorelq:
        cmd = cmd+' --ignorelq'
    if options.refmin > 0:
        cmd = cmd+' --refmin '+str(options.refmin)
    if options.tsmooth > 0:
        cmd = cmd+' --tsmooth '+str(options.tsmooth)
    if options.gsmooth > 0:
        cmd = cmd+' --gsmooth '+str(options.gsmooth)
    if options.param2d is not None:
        cmd = cmd+' '+options.param2d
    if not outsurf == "":
        cmd = cmd+' --outmesh '+outsurf
    run_cmd(cmd,'shapeDNA-tria failed?') 


if __name__=="__main__":
    # Command Line options and error checking done here
    options = options_parse()

    my_print(options.label)
    my_print(options.surf)

    if options.asegid is not None:
        surf = get_aseg_surf(options.sdir,options.sid,options.asegid,options.outsurf)
        outsurf = surf
    elif options.label is not None:
        surf = get_label_surf(options.sdir,options.sid,options.label,options.surf,options.source,options.outsurf)
        outsurf = surf
    elif options.aparcid is not None:
        surf = get_aparc_surf(options.sdir,options.sid,options.surf,options.aparcid,options.outsurf)    
        outsurf = surf
    elif options.surf is not None:
        surf = get_surf_surf(options.sdir,options.sid,options.surf)
        outsurf = options.outsurf
    #my_print(surf)
    #sys.exit(0)
    if surf is None:
        my_print('ERROR: no surface was created/selected?')
        sys.exit(1)

    if options.dotet:
        #convert to tetmesh
        get_tetmesh(surf,options.outtet,options.fixiter)
        #run shapedna tetra
        run_shapeDNAtetra(options.outtet,options.outev,options)
    else:
        run_shapeDNAtria(surf,options.outev,outsurf,options)

    # always exit with 0 exit code
    sys.exit(0)



from nipype.interfaces.freesurfer import MRIMarchingCubes
from neuroshape.utils.checks import is_string_like
#import gmsh
from lapy import TriaMesh
import warnings
from collections import OrderedDict
import scipy.optimize as optimize

#gmsh.initialize()

import numpy as np
"""
Read and write geometry into different formats

    - Runs mri_mc from Freesurfer to create 2d surface
    - Projects 2d surface to gmsh and creates a tetrahedral mesh for LaPy
    - Writes out geometry in tetrahedral format or in Freesurfer binary
    - Writes out label files

Code was taken from nibabel.freesurfer package (https://github.com/nipy/nibabel/blob/master/nibabel/freesurfer/io.py).
"""

def mri_mc(in_file, label_value):
    """
    Runs class nipype.interfaces.freesurfer.MRIMarchingCubes()

    Parameters
    ----------
    in_file : str
        filename of label volume
    label_value : int
        label or mask value to compute marching cubes algorithm on
    out_file : str
        filename of output

    Returns
    -------
    str
        Name of the output file
        
    """
    
    mc = MRIMarchingCubes()
    mc.inputs.in_file = in_file
    mc.inputs.label_value = label_value
    mc.run()
    
    return mc.output_spec.out_file

def combine_geometry(in_file, label_values):
    geom = np.array()
    for label in label_values:
        if not type(label) == int:
            return ValueError('incorrect format for label inputs')
        
        out_file = mri_mc(in_file, label)
        geom = np.append(read_geometry(out_file))
    
    tria = TriaMesh(geom)
    
    return tria

def gaussian(x, amplitude, mean, stddev):
    return amplitude * np.exp(- ((x - mean) ** 2) / (2 * (stddev ** 2)))

def estimate_fwhm(image):
    # Find the maximum value and its position in the image
    max_val = np.max(image)
    max_pos = np.argmax(image)
    max_row, max_col = np.unravel_index(max_pos, image.shape)
    
    # Create a 1d profile along the maximum row and column
    row_profile = image[max_row, :]
    col_profile = image[:, max_col]
    
    options = {'maxfev' : np.prod(image.shape)}
    
    # Fit a gaussian curve to the row profile
    row_x = np.arange(len(row_profile))
    row_params, _ = optimize.curve_fit(gaussian, row_x, row_profile, p0=[max_val, max_col, 1], **options)
    
    # Fit a gaussian curve to the column profile
    col_x = np.arange(len(col_profile))
    col_params, _ = optimize.curve_fit(gaussian, col_x, col_profile, p0=[max_val, max_row, 1], **options)
    
    # Calculate the FWHM as 2.355 times the std dev of the Gaussian
    row_fwhm = 2.355 * row_params[2]
    col_fwhm = 2.355 * col_params[2]
    
    # take the average
    fwhm = (row_fwhm + col_fwhm) / 2
    
    return fwhm
    
def resel_count(image, fwhm):
    # returns resel count of image given FWHM
    return np.prod(image.shape / fwhm)    
    
# def write_geometry(model, outfile):
#     if not is_string_like(outfile):
#         return ValueError('incorrect format for filename output')
#     if not outfile[:-4] == ".msh":
#         outfile += ".msh"
    
#     gmsh.write(outfile)
#     gmsh.finalize()
    
#     return print(f'mesh file saved to {outfile}')
        
def convert_geometry(in_file, label_value, out_file=None):
    """
    Converts label or mask file to triangular mesh of class lapy.TriaMesh()

    Parameters
    ----------
    in_file : str
        filename of label volume
    label_value : int
        label or mask value to compute marching cubes algorithm on
    out_file : str, optional
        filename of output

    Returns
    -------
    tria : lapy.TriaMesh class

    Raises
    ------
    ValueError : inputs are incorrectly formatted

    """
    if not is_string_like(in_file) or not is_string_like(out_file):
        return ValueError(f'expected str and str, got {type(in_file)}, {type(out_file)}')
    if not type(label_value) == int:
        return ValueError(f'label must be an integer, got {type(label_value)}')
    
    if out_file is not None:
        mri_mc(in_file, label_value, out_file)
    else:
        out_file = mri_mc(in_file, label_value)
        
    coords, faces = read_geometry(out_file)
    
    tria = TriaMesh(coords, faces)
    
    return tria
    
def _fread3(fobj):
    """Read a 3-byte int from an open binary file object
    Parameters
    ----------
    fobj : file
        File descriptor
    Returns
    -------
    n : int
        A 3 byte int
    """
    b1, b2, b3 = np.fromfile(fobj, ">u1", 3)
    return (b1 << 16) + (b2 << 8) + b3


def _read_volume_info(fobj):
    """Helper for reading the footer from a surface file."""
    volume_info = OrderedDict()
    head = np.fromfile(fobj, ">i4", 1)
    if not np.array_equal(head, [20]):  # Read two bytes more
        head = np.concatenate([head, np.fromfile(fobj, ">i4", 2)])
        if not np.array_equal(head, [2, 0, 20]) and not np.array_equal(
            head, [2, 1, 20]
        ):
            warnings.warn("Unknown extension code.")
            return volume_info
        head = [2, 0, 20]

    volume_info["head"] = head
    for key in [
        "valid",
        "filename",
        "volume",
        "voxelsize",
        "xras",
        "yras",
        "zras",
        "cras",
    ]:
        pair = fobj.readline().decode("utf-8").split("=")
        if pair[0].strip() != key or len(pair) != 2:
            raise IOError("Error parsing volume info.")
        if key in ("valid", "filename"):
            volume_info[key] = pair[1].strip()
        elif key == "volume":
            volume_info[key] = np.array(pair[1].split()).astype(int)
        else:
            volume_info[key] = np.array(pair[1].split()).astype(float)
    # Ignore the rest
    return volume_info    
    
def read_geometry(filepath, read_metadata=False, read_stamp=False):
    """Read a triangular format Freesurfer surface mesh.
    Parameters
    ----------
    filepath : str
        Path to surface file.
    read_metadata : bool, optional
        If True, read and return metadata as key-value pairs.
        Valid keys:
        * 'head' : array of int
        * 'valid' : str
        * 'filename' : str
        * 'volume' : array of int, shape (3,)
        * 'voxelsize' : array of float, shape (3,)
        * 'xras' : array of float, shape (3,)
        * 'yras' : array of float, shape (3,)
        * 'zras' : array of float, shape (3,)
        * 'cras' : array of float, shape (3,)
    read_stamp : bool, optional
        Return the comment from the file
    Returns
    -------
    coords : numpy array
        nvtx x 3 array of vertex (x, y, z) coordinates.
    faces : numpy array
        nfaces x 3 array of defining mesh triangles.
    volume_info : OrderedDict
        Returned only if `read_metadata` is True.  Key-value pairs found in the
        geometry file.
    create_stamp : str
        Returned only if `read_stamp` is True.  The comment added by the
        program that saved the file.
    """
    volume_info = OrderedDict()

    TRIANGLE_MAGIC = 16777214

    with open(filepath, "rb") as fobj:
        magic = _fread3(fobj)

        if magic == TRIANGLE_MAGIC:  # Triangle file
            create_stamp = fobj.readline().rstrip(b"\n").decode("utf-8")
            test_dev = fobj.peek(1)[:1]
            if test_dev == b"\n":
                fobj.readline()
            vnum = np.fromfile(fobj, ">i4", 1)[0]
            fnum = np.fromfile(fobj, ">i4", 1)[0]
            coords = np.fromfile(fobj, ">f4", vnum * 3).reshape(vnum, 3)
            faces = np.fromfile(fobj, ">i4", fnum * 3).reshape(fnum, 3)

            if read_metadata:
                volume_info = _read_volume_info(fobj)
        else:
            raise ValueError(
                "File does not appear to be a Freesurfer surface (triangle file)"
            )

    coords = coords.astype(float)  # XXX: due to mayavi bug on mac 32bits

    ret = (coords, faces)
    if read_metadata:
        if len(volume_info) == 0:
            warnings.warn("No volume information contained in the file")
        ret += (volume_info,)
    if read_stamp:
        ret += (create_stamp,)

    return ret

def write_label(filepath, vertices, values=None):
    """
    Write Freesurfer label data `values` to filepath `filepath`

    Parameters
    ----------
    filepath : str
        String containing path to label file to be written
    vertices : ndarray, shape (n_vertices, 3)
        Coordinates of each vertex
    values : optional, shape (n_vertices,)
        Array of scalar data for each vertex. The default is None.
    """
    
    if values is not None:
        vector = np.asarray(values)
        vnum = np.prod(vector.shape)
        if vector.shape not in ((vnum,), (vnum, 1), (1, vnum), (vnum, 1, 1)):
            raise ValueError('Invalid shape: argument values must be a vector')
            
    else:
        vector = np.zeros(vnum, dtype=np.float32)
    
    start_line = '#!ascii label  , from subject  vox2ras=TkReg\n'
    magic_number = vnum + 6000
    
    with open(filepath, 'w') as fobj:
        fobj.write(start_line)
        fobj.write(f'{magic_number}\n')
        
        #now write vertex and label array
        label_array = np.vstack((np.array(range(vnum)), vertices.T, vector.T)).T.astype('>f4')
        np.savetxt(fobj, label_array, fmt=['%i','%f','%f','%f','%f'])


# -*- coding: utf-8 -*-
"""
Utility functions
Taken from neuromaps/utils (c) netneurolab 2022
"""

import os

def check_fs_subjid(subject_id, subjects_dir=None):
    """
    Checks that `subject_id` exists in provided FreeSurfer `subjects_dir`

    Parameters
    ----------
    subject_id : str
        FreeSurfer subject ID
    subjects_dir : str, optional
        Path to FreeSurfer subject directory. If not set, will inherit from
        the environmental variable $SUBJECTS_DIR. Default: None

    Returns
    -------
    subject_id : str
        FreeSurfer subject ID, as provided
    subjects_dir : str
        Full filepath to `subjects_dir`

    Raises
    ------
    FileNotFoundError
    """

    # check inputs for subjects_dir and subject_id
    if subjects_dir is None or not os.path.isdir(subjects_dir):
        try:
            subjects_dir = os.environ['SUBJECTS_DIR']
        except KeyError:
            subjects_dir = os.getcwd()
    else:
        subjects_dir = os.path.abspath(subjects_dir)

    subjdir = os.path.join(subjects_dir, subject_id)
    if not os.path.isdir(subjdir):
        raise FileNotFoundError('Cannot find specified subject id {} in '
                                'provided subject directory {}.'
                                .format(subject_id, subjects_dir))

    return subject_id, subjects_dir


from scipy.interpolate import griddata
import numpy as np
from ..utils.concavehull import ConcaveHull
from shapely.geometry import Point

def grid(x, vertices, tol=20, spacing=0.5, ch=None):
    """
    Interpolates surface onto grid for wavelet transform
    
    Parameters
    ----------
    x : (N,) np.ndarray
        Data values at each surface vertex
    vertices : (N,3) np.ndarray
        Vertices of gifti object
    sf : float
        Shrink factor, default 0.7
    spacing : float
        Target grid spacing in mm, default 0.5
    
    Returns
    -------
    interpolated_data : (N,M) np.ndarray
        Interpolated data on the grid 'xi'
    gridX : (N,) np.ndarray
        Interpolated x coordinates of vertices on the grid
    gridY : (N,) np.ndarray
        Interpolated y coordinates of vertices on the grid   
    
    Notes
    -----
    Data values outside of convex hull are left as 'np.nan'
    """
    
    c = x
    x = vertices[:, 0]
    y = vertices[:, 1]
    gridX, gridY = np.mgrid[min(x):max(x):spacing, min(y):max(y):spacing]
    
    unbounded_data = griddata((x, y), c, (gridX, gridY), method='nearest', fill_value=0)
    
    # if not ch:
    #     ch = ConcaveHull()
    #     ch.loadpoints(vertices[:, :2])
    #     ch.calculatehull(tol=tol)
    
    bounded_data = np.zeros(unbounded_data.shape)*np.nan
    for i in np.arange(len(unbounded_data)):
        for j in np.arange(len(unbounded_data[i])):
            if ch.boundary.contains(Point(gridX[i,j], gridY[i,j])):
                bounded_data[i,j] = unbounded_data[i,j]
            else:
                continue
            
    bounded_data[np.isnan(bounded_data)] = 0
        
    return bounded_data, gridX, gridY

def mesh(interpolated_data, vertices, gridX, gridY):
    """
    Backprojects grid from 'grid()' to surface
    
    Parameters
    ----------
    interpolated_data : (N,M) np.ndarray
        Output from 'grid()', interpolated grid to perform wavelet transforms
    vertices : (N,3) np.ndarray
        Vertices of gifti object
    gridX : (N,N) np.ndarray
        Interpolated X coordinates
    gridY : (N,N) np.ndarray
        Interpolated Y coordinates
    
    Returns
    -------
    uninterpolated_data : (N,)
        Uninterpolated data output from backprojecting grid onto vertex coordinates
    
    Notes
    -----
    Data values outside of convex hull are left as 'np.nan'
    """
    c = np.matrix.flatten(interpolated_data)    
    x = vertices[:, 0]
    y = vertices[:, 1]
    uninterpolated_data = griddata((gridX.flatten(), gridY.flatten()), c, (x, y), method='linear')
    
    
    return uninterpolated_data



from lapy import TriaMesh
from lapy.ShapeDNA import compute_shapedna
import numpy as np
from joblib import Parallel, delayed
from .utils.dataio import dataio
from pathlib import Path

__all__ = ['PolyLBO']

class PolyLBO:
    """
    Description
    -----------
    Wraps the Laplace-Beltrami Operator as implemented in ShapeDNA, see:
        https://en.wikipedia.org/wiki/Laplace%E2%80%93Beltrami_operator
        and:
        http://reuter.mit.edu/software/shapedna/
        
    Runs the Laplace-Beltrami calculation in parallel (as implemented in
    .__call__() and ._call_method()) and outputs an np.ndarray of size (n,J,N) 
    of "eigenmodes" for J=200 as in Koussis et al. and where N is the number
    of vertices and n is the number of files given as input to `x`.
    
    Dependencies
    ------------
        'nibabel', 
        'lapy', 
        'nilearn', 
        'numpy', 
        'scipy', 
        'scikit-sparse',
        'neuromaps', 
        'nipype'
        
    Inputs
    ------
        See .utils.dataio()
        x : list_like or array_like or string_like or file_like : 
            list-like or array-like or string-like or file-like
            
            if array_like : ((n_vertices,3),(n_faces,3)) of np.ndarrays
            
            if string_like : string containing path of file to be loaded into
            memory. Must have .gii extension, other extensions (e.g., .vtk
            meshes) are future implementations.
            
            if file-like : string containing paths of files to be loaded into
            memory. Must have .gii extension, other extensions 
            (e.g., .vtk meshes) are future implementations.
            
        Vn : int, optional (default 2)
            Which eigenvalue to compute group contrast on. Vn cannot be greater
            than eigs.
            
        eigs : int, optional (default 2)
            Number of eigenvalues to compute using ShapeDNA.
            
        n_jobs : int, optional (default 1)
            Number of jobs to use for .compute_eigenmap()
                    
    Outputs
    -------
        See .utils.dataio()
        np.ndarray or list_like : np.ndarray or list-like
            Output type defined by input.
            
    Raises
    ------
    ValueError : input is invalid
    RuntimeError : too many files in `x`
    TypeError : x is not one of the types listed above
        
    """

    def __init__(self, eigs=200, n_jobs=1):
        self._n_jobs = n_jobs
        self._eigs = eigs + 1 #to account for the zeroth mode
        
        #self._x = [dataio(i) for i in x]
        
        #n = len(self._x)
        
        # if not n > 0:
        #     raise RuntimeError("expected filename, ndarray, string array, or .txt file, got empty")
        
    def __call__(self, x):
        """
        Generate LBO(s) for filenames in native space listed in `x`.

        Parameters
        ----------
        x : list-like or array-like or string-like or file-like

        Returns
        -------
        (n,J,N) np.ndarray
            Generated LBO eigenmap(s) of n surfaces by N = `n_vertices`

        """
        
        if type(x) == list:
            if isinstance(x[0][0], np.ndarray):
                x_ = x
            else:
                x_ = [dataio(i) for i in x]
        elif type(x) == tuple:
            try:
                return np.asarray(self.compute_eigenmaps(x)).squeeze()
            except AttributeError:
                ValueError("inputs are invalid")
        elif type(x) == str:
            x_ = dataio(x)
            return np.asarray(self.compute_eigenmaps(x=x_)).squeeze()
        elif Path(x).suffix == ".txt":
            txt = dataio(x)
            x_ = dataio(txt)
        else:
            raise TypeError("expected list-like or array-like or string-like or file-like, got {}".format(type(x)))
        
        eigmaps = Parallel(
            self._n_jobs, prefer='threads')(
                delayed(self._call_method)(x=i) for i in x_
               )
                
       
        return eigmaps
    
    def _call_method(self, x=None, hemi=None):
        """
        Subfunction used by .__call__() for parallelization purposes

        """
        #generate eigenmaps
        eigs = self._eigs
        eigvecs = self.compute_eigenmaps(x, eigs=eigs)
        
        return eigvecs.squeeze()
    
    def compute_eigenmaps(self, x, eigs):
        """
        Computes Laplace-Beltrami operator as implemented in ShapeDNA

        Parameters
        ----------
        x : (2,(N, 3)) tuple of np.ndarray
            Brain map
        eigs : int
            Number of eigenvalues to calculate on the surface `x`. 

        Returns
        -------
        (N, eigs) np.ndarray
            Eigenmaps of LBO eigenvalues of n_vertices with 
            number of columns = len(eigs)

        """
        coords, faces = x
        tria = TriaMesh(coords, faces)
        ev = compute_shapedna(tria, k=eigs)
        #eigvecs = ev['Eigenvectors'][:,Vn-1]
        #eigvecs = eigvecs/np.linalg.norm(eigvecs, ord=np.inf)
        
        
        return ev['Eigenvectors'][:,1:]
    


import numpy as np

def check_map(x):
    """
    Check that brain map conforms to expectations.

    Parameters
    ----------
    x : np.ndarray
        Brain map

    Returns
    -------
    None

    Raises
    ------
    TypeError : `x` is not a np.ndarray object
    ValueError : `x` is not one-dimensional

    """
    if not isinstance(x, np.ndarray):
        e = "Brain map must be array-like\n"
        e += "got type {}".format(type(x))
        raise TypeError(e)
    if x.ndim != 1:
        e = "Brain map must be one-dimensional\n"
        e += "got shape {}".format(x.shape)
        raise ValueError(e)


import numpy as np
from neuromaps.stats import compare_images
from numpy.random import permutation as perm
from numpy.random import randint
from .swap_single_row import swap_single_row
from lapy.TriaMesh import TriaMesh
import nibabel as nib

"""
Eigenmode helper functions (C) Systems Neuroscience Newcastle &
Nikitas C. Koussis 2023
"""

def maximise_recon_metric(eigs, y, metric='corr'):
    """
    Takes a set of eigenmodes `eigs` and a single eigenmode `y` and swaps 
    eigenmodes within an eigengroup to maximise `corr` in a
    reconstructed map of `y`. Other metrics are not implemented.

    Parameters
    ----------
    eigs : (N,M) np.ndarray
        Eigenmode array to swap eigenmodes within eigengroups with N 
        eigenmodes and M vertices.
    y : (M,) or (M,1) or (1,M) np.ndarray
        Functional gradient map
    metric : str, optional
        Metric to maximise. The default is 'corr'.

    Returns
    -------
    new_eigs : (N,M) np.ndarray
        Eigenmode array that maximises correlation within eigengroups.
    metric_out : float
        Maximized metric value
        
    Raises
    ------
    NotImplementedError : `metric` is not in the implemented classes

    """
        
    # swap within groups to maximise metric
    if metric != 'corr':
        raise NotImplementedError('{} is not implemented yet'.format(str(metric)))
        
    # check if y and eigs in proper orientation
    if not isinstance(eigs, np.ndarray) or not isinstance(y, np.ndarray):
        raise TypeError('Eigenmodes and functional maps must be array-like, got type {}, and {}'.format(type(eigs),type(y)))
    if eigs.ndim != 2 or y.ndim != 1:
        raise ValueError('Eigenmodes must be 2-D and functional map must be 1-D, got {}D and {}D'.format(eigs.ndim, y.ndim))
    if eigs.shape[1] != y.shape[0]:
        if eigs.T.shape[1] == y.shape[0]:
            eigs = eigs.T
        else:
            raise RuntimeError('Eigenmodes and functional maps must be able to be matrix multiplied, fix')
    
    groups = _get_eigengroups(eigs)
    
    # iterative swapping of eigenmodes within eigengroup to maximise metric   
    new_eigs = find_optimum_eigengroups(eigs, y, groups)
    
    return new_eigs


def _get_eigengroups(eigs):
    """
    Helper function to find eigengroups
    """
    lam = eigs.shape[0] # number of eigenmodes
    l = np.floor((lam-1)/2).astype(int)    
    if lam == 1:
        return np.asarray([0])
    if lam == 2:
        groups = [np.zeros(1).astype(int)]
        groups.append(np.ones(1).astype(int))
        return groups
    
    groups = []
    ii = 0
    i = 0
    for g in range(l+1):
        ii += 2*g + 1
        if ii >= lam:
            groups.append(np.arange(i,lam))
            break
        groups.append(np.arange(i,ii))
        i = ii
    
    return groups


def find_optimum_eigengroups(eigs, y, groups, previous_corr=0., tol=0.00001):
    #copy original array
    eigs_ = eigs
    if len(groups) == 2:
        if len(groups[0]) < 2:
            return eigs_
    
    eigs_swapped = np.vstack(swap_single_row(eigs_[groups[i]]) for i in range(len(groups)))
    next_betas = np.matmul(eigs_swapped, y)
    next_recon = np.matmul(next_betas.T, eigs_swapped).reshape(-1,)
    next_corr = compare_images(y, next_recon)
    
    try:
        if (next_corr - previous_corr) > tol:
            return eigs_
    except:
        return eigs_
    eigs_ = eigs_swapped
    previous_corr = next_corr
    

def reconstruct_surface(surface, eigenmodes, n=100, normalize='area', norm_factor=1, method=None):
    """
    Reconstruct a surface of `n_vertices` given a set of eigenmodes

    Parameters
    ----------
    surface : nib.GiftiImage type
        Surface to reconstruct of `n_vertices` (must be a single hemisphere or a single surface)
    coeffs : np.ndarray of shape (n_vertices, 3)
        Coefficients of eigen decomposition
    eigenmodes : np.ndarray of shape (n_vertices, M)
        Eigenmodes of `n_vertices` by number of eigenvalues M
    n : int (default 100)
        Number of eigenmodes to use for reconstruction
    normalize : str, optional
        How to normalize the new surface. The default is 'area'.
        
        Accepted types:
            'constant' : normalize by a constant factor (default is 1^(1/3))
            'number' : normalize by the number of vertices
            'volume' : normalize by the volume of the reconstructed surface
            'area' : normalize by the area of the new vertices
            
    method : str, optional
        method of calculation of coefficients: 'matrix', 'matrix_separate', 
        'regression'
        The default is 'matrix'

    Returns
    -------
    new_surface : nib.GiftiImage type
        Reconstructed surface

    """
    
    # initialize new vertices
    new_vertices = np.zeros_like(eigenmodes)
    
    # find coeffs
    if method is not None:
        coeffs = eigen_decomposition(surface.darrays[0].data, eigenmodes, method=method)
    else:
        coeffs = eigen_decomposition(surface.darrays[0].data, eigenmodes)
        
    
    # partition coeffs into x, y, z
    coeffs = coeffs.T
    coeffs_x, coeffs_y, coeffs_z = coeffs
    
    new_vertices[:, 0] = eigenmodes[:, :n] @ coeffs_x
    new_vertices[:, 1] = eigenmodes[:, :n] @ coeffs_y
    new_vertices[:, 2] = eigenmodes[:, :n] @ coeffs_z

    faces = surface.darrays[1].data
    
    # normalize vertices
    if normalize:
        tria = TriaMesh(new_vertices, faces)
        tria_norm = tria
        if normalize == 'number':
            number = np.sum(tria.v.shape)
            tria_norm.v = tria.v/(number ** (1/3))
            
        elif normalize == 'volume':
            volume = tria.volume()
            tria_norm.v = tria.v/(volume ** (1/3))
            
        elif normalize == 'constant':
            tria_norm.v = tria.v/(norm_factor ** (1/3))
        
        elif normalize == 'areas':
            areas = tria.vertex_areas()
            tria_norm.v = tria.v/(areas ** (1/3))
        
        else:
            pass
        new_vertices = tria_norm.v
        
    new_surface = nib.GiftiImage()
    new_surface.add_gifti_data_array(nib.gifti.gifti.GiftiDataArray((new_vertices, faces)))
    
    return new_surface

    
def eigen_decomposition(data, eigenmodes, method='matrix'):
    """
    Decompose data using eigenmodes and calculate the coefficient of 
    contribution of each vector
    
    Parameters:
    -----------
    data : np.ndarray of shape (n_vertices, 3)
        N = number of vertices, P = columns of independent data
    eigenmodes : np.ndarray of shape (n_vertices, M)
        N = number of vertices, M = number of eigenmodes
    method : string
        method of calculation of coefficients: 'matrix', 'matrix_separate', 
        'regression'
    
    Returns:
    -------
    coeffs : numpy array of shape (N, 3)
     coefficient values
    
    """
    
    N, P = data.shape
    _, M = eigenmodes.shape
    
    if method == 'matrix':
        coeffs = np.linalg.solve(eigenmodes.T @ eigenmodes, eigenmodes. T @ data)
    
    elif method == 'matrix-separate':
        coeffs = np.zeros((N, P))
        for p in range(P):
            coeffs[:, p] = np.linalg.solve(eigenmodes.T @ eigenmodes, eigenmodes.T @ data[:, p])
            
    elif method == 'regression':
        coeffs = np.zeros((N, P))
        for p in range(P):
            coeffs[:, p] = np.linalg.lstsq(np.hstack((eigenmodes, np.ones((M, 1)))), data[:, p], rcond=None)[0][:-1]
                
    return coeffs
    
    
def compute_axes_ellipsoid(eigenvalues):
    """
    Compute the axes of an ellipsoid given the eigenmodes.
    Data is a 2D numpy array where each row is a data point and each column represents a dimension.
    """    
    return np.sqrt(eigenvalues)
    

def transform_to_spheroid(eigenvalues, eigenmodes):
    """
    Transform the eigenmodes to a spheroid space
    """
    ellipsoid_axes = compute_axes_ellipsoid(eigenvalues)
    ellipsoid_axes = ellipsoid_axes.reshape(-1, 1)
    
    spheroid_eigenmodes = eigenmodes / ellipsoid_axes
    
    return spheroid_eigenmodes
    
    
def transform_to_ellipsoid(eigenvalues, eigenmodes):
    """
    Transform the eigenmodes in spheroid space back to ellipsoid by stretching
    """
    
    ellipsoid_axes = compute_axes_ellipsoid(eigenvalues)
    ellipsoid_axes = ellipsoid_axes.reshape(-1, 1)
    
    ellipsoid_eigenmodes = eigenmodes * ellipsoid_axes
    
    return ellipsoid_eigenmodes


def resample_spheroid(spheroid_eigenmodes, angles=None):
    """
    Resample the N-D spheroid generated by the N orthogonal unit modes

    """
    # radius = 1 on the unit hypersphere
    r = 1
    # ensure the eigenmodes are normalized on the unit spheroid
    spheroid_eigenmodes = spheroid_eigenmodes / np.linalg.norm(spheroid_eigenmodes, axis=0)
    
    # initialize the new points p
    p = r * spheroid_eigenmodes
    
    # check if angles are input or not
    if angles:
        assert angles.shape[0] == spheroid_eigenmodes.shape[0] - 1, "The number of angles should be one less than the number of basis modes"
        angles = angles
    else:
        angles = np.random.random_sample(size=spheroid_eigenmodes.shape[0]) * 2 * np.pi
        
    # Compute the coordinates for new points p
    print("Computing the coordinates for each dimension")
    for i in range(1, spheroid_eigenmodes.shape[0]):
        r_i = r
        for j in range(i):
            if np.mod(i, 2) == 1: #ODD
                if angles[j] > np.pi:
                    angles[j] = (angles[j] % np.pi)*np.pi
            else: #EVEN
                if angles[j] > 2*np.pi: 
                    angles[j] = (angles[j] % 2*np.pi)*2*np.pi
                r_i *= np.sin(angles[j])
        if i < spheroid_eigenmodes.shape[0] - 1:
            r_i *= np.cos(angles[j-1])
        
        p += r_i * spheroid_eigenmodes[i]
        
    # find the unit modes that describe the points p
    new_modes = p / np.linalg.norm(p)
    
    # ensure that the unit modes are orthonormal (QR decomposition)
    print("Ensuring the new modes are orthonormal")
    new_modes = np.linalg.qr(new_modes, mode='reduced')
    
    return new_modes  


"""
Utility functions
Taken from neuromaps/utils (c) netneurolab 2022
"""

import os
from pathlib import Path
import tempfile


def tmpname(suffix, prefix=None, directory=None):
    """
    Little helper function because :man_shrugging:

    Parameters
    ----------
    suffix : str
        Suffix of created filename

    Returns
    -------
    fn : str
        Temporary filename; user is responsible for deletion
    """

    fd, fn = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=directory)
    os.close(fd)

    return Path(fn)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Calculate the connectopic Laplacian of an ROI given some fmri data

Usual usage:
    $ data_input_filename=fmri_input.nii.gz \
    $ data_roi_filename=roi.nii.gz \
    $ mask_filename=mask.nii.gz \
    $ data_output_filename=gradients.nii.gz \
    $ num_gradients=20 \
    $ fwhm=6 \
    $ python connectopic_laplacian.py ${data_input_filename} ${data_roi_filename} ${mask_filename} ${data_output_filename} -N ${num_gradients} --smoothing ${fwhm} --filtering

@author: Nikitas C. Koussis, Systems Neuroscience Group Newcastle, 2023
"""

import nibabel as nib
import numpy as np
from scipy.signal import detrend
from scipy.ndimage import gaussian_filter
from scipy.spatial.distance import pdist, squareform
from scipy.sparse.csgraph import laplacian
from numpy.linalg import svd
from numpy.linalg import eigh
import os
from argparse import ArgumentParser
from nilearn import image, masking
from subprocess import Popen
from neuroshape.eta import eta_squared
from neuroshape.euler import euler_threshold
from os.path import split

os_path = dict(os.environ).get('PATH')

global matlabpath
path_list = os_path.split(sep=':')
for item in path_list:
    global matlabpath
    if 'matlab' in item:
        matlabpath = item
    if 'MATLAB' in item:
        matlabpath = item
    # else:
    #     raise RuntimeError("MATLAB is not installed or has not been sourced on the variable $PATH")

image_types = [
    '.nii',
    '.nii.gz',
]

def smooth(input_filename, fwhm):
    # load data
    img_input = image.load_img(input_filename)
    data_input = img_input.get_fdata()
    # number of volumes
    T = img_input.shape[-1]
    
    # sigma
    sigma = fwhm / np.sqrt(8 * np.log(2))
    
    # initialize smoothed image
    data_smoothed = np.zeros_like(data_input)
    
    for vol in range(T):
        data_smoothed[:,:,:,vol] = gaussian_filter(data_input[:,:,:,vol], sigma=sigma)
        
    # save smoothed nifti
    img = nib.Nifti1Image(data_smoothed, img_input.affine, header=img_input.header)
    output_filename = input_filename.replace('.nii','_smoothed.nii')
    print("Saving smoothed file to {}".format(output_filename))
    nib.save(img, output_filename)
    
    return output_filename

def wishart(input_filename, mask_filename):
    # write matlab script
    script_file = os.path.join('/tmp', 'wishart_run.m')
    folder, *_ = split(__file__)
    with open(script_file, 'w') as file:
        file.write(f"""
addpath {folder}/functions/wishart
addpath {folder}/functions

DEMDT = 1;
VN = 1;
Iterate = 2;
NDist = 1;

[~, data] = read('{input_filename}');

[~, gm_msk] = read('{mask_filename}');
ind_gm = find(gm_msk);

[folder, basename, ~] = fileparts('{input_filename}');

T = size(data,4);
x = zeros(T, length(ind_gm));

for i=1:T
    tmp = data(:,:,:,i);
    x(i,:) = tmp(ind_gm);
end

Out = icaDim(x', DEMDT, VN, Iterate, NDist);

x = Out.data';

x = detrend(x, 'constant');
x = x./repmat(std(x), T, 1);

new_data = zeros(size(gm_msk,1), size(gm_msk,2), size(gm_msk,3), T);
[xx, yy, zz] = ind2sub(size(gm_msk), ind_gm);

x = x';

for i = 1:length(ind_gm)
    new_data(xx(i), yy(i), zz(i), :) = x(i,:);
end

mat2nii(new_data, [folder, '/', basename, '_filtered.nii'], size(new_data), 32, '{mask_filename}');
                   """)
    
    # run matlab at command line
    matpath = [f'{matlabpath}']
    options = ['-nosplash', '-nodesktop', '-r']
    command = ["run('{0}'); exit;".format(script_file)]
    
    p = Popen(matpath + options + command)
    stderr, stdout = p.communicate()
        
    if stderr is not None:
        raise RuntimeWarning(stderr)
        print('Could not perform filtering, check logs')
        return input_filename
    
    os.system(f'rm -f {script_file}')  
    output_filename = input_filename.replace('.nii','_filtered.nii')
    print("Saving filtered file to {}".format(output_filename))
    
    return output_filename

def normalize_data(data):
    data_normalized = np.subtract(data, np.mean(data, axis=0))
    data_normalized = np.divide(data_normalized, np.std(data_normalized, axis=0))
    
    return data_normalized

def compute_similarity(img_input, img_roi, img_mask):
    data_msk = masking.apply_mask(img_input, img_mask)
    print('Normalizing input timeseries')
    data_msk = normalize_data(data_msk)
    input_img = masking.unmask(data_msk, img_mask)
    
    data_ins = masking.apply_mask(input_img, img_roi)
    
    T = data_ins.shape[0]
    
    print('Running singular value decomposition on input timeseries')
    U, S, _ = svd(data_msk, full_matrices=False)
    
    a = U.dot(np.diag(S))
    a = a[:, :-1]
    
    a = normalize_data(a)
        
    print('Computing similarity matrix by eta squared coefficient')
        
    c = data_ins.T.dot(a)
    c = c / T
    zpc = np.arctanh(c)
    zpc = zpc[:,np.all(~np.isnan(zpc), axis=0)]
    
    smat = eta_squared(zpc)
    
    return smat

def thresh(w, s):
    # # find FWHM of matrix
    # fwhm = estimate_fwhm(w)
    # # calculate resel count using FWHM
    num_resels = 256 #resel_count(w, fwhm) DO 256 FOR NOW
    # compute optimal threshold based on euler characteristic
    threshold = euler_threshold(w, num_resels)
    w_thresh = w <= threshold
    binary = w_thresh > 0.
    s_thresh = np.zeros_like(s)
    s_thresh[binary] = s[binary]
    n_edges = np.sum(np.size(s_thresh > 0.))
    n_nodes = s_thresh.shape
    density = 2.0*n_edges / np.prod(n_nodes)
    
    return s_thresh, density

def calc_LaplacianMatrix(s, num_gradients):
    # distance mapping
    w = squareform(pdist(s))
    
    # threshold matrix
    print('Thresholding using maximum Euler characteristic at 256 resels')
    s_thresh, density = thresh(w, s)
    
    print('Density of similarity graph needed to remain connected: {:.2f}%'.format(density))
    
    print('Computing Laplacian')
    L = laplacian(s_thresh)
    
    return L
    
def calc_gradients(img_input, img_roi, img_mask, num_gradients=2):
    smat = compute_similarity(img_input, img_roi, img_mask)
    
    L = calc_LaplacianMatrix(smat, num_gradients)
    
    evals, egrads = eigh(L)
    
    # exclude the first eigenvalue, return only the non-zero `num_gradients` last gradients
    egrads = egrads[:, 1:num_gradients+1]
    evals = evals[1:num_gradients+1]
    
    # Z-transform eigenvectors
    egrads = normalize_data(egrads)
    
    # make positive
    egrads = egrads - np.min(egrads)
    
    
    return evals, egrads

def cortical_projection(img_input, img_roi, img_mask, egrads):
    data_msk = masking.apply_mask(img_input, img_mask)
    data_msk = normalize_data(data_msk)
    input_img = masking.unmask(data_msk, img_mask)
    
    data_ins = masking.apply_mask(input_img, img_roi)
    roi_index = [data_msk == data_ins]
    
    T = data_ins.shape[0]
    
    u, s, v = svd(data_msk, full_matrices=False)
    
    a = u * s
    a = a[:,:-1]
    
    a = normalize_data(a)
        
    c = np.matmul(data_ins.T, a)
    c = np.divide(c, T)
    
    cc = np.hstack((c, np.zeros(c.shape[0]))) * v.T
    
    index_max_cc = find_roi_index(cc, data_ins, roi_index)
    
    ind_msk = np.where(data_msk > 0.)
    eigvec_proj = np.zeros((img_mask.shape[0], img_mask.shape[1], img_mask.shape[2], egrads.shape[1]))
    
    for grad in range(egrads.shape[1]):
        eig = egrads[:, grad]
        eig_proj = eig[index_max_cc]
        xx, yy, zz = np.unravel_index(ind_msk, img_mask.shape)
        for i in data_msk.shape[0]:
            if not np.in1d(ind_msk, roi_index):
                eigvec_proj[xx[i], yy[i], zz[i], grad] = eig_proj[i]
    
    return eigvec_proj
            
    
def find_roi_index(cc, roi_data, roi_index):
    print("Finding cortex-to-roi projection by maximum correlation")
    
    max_cc = np.max(cc[roi_index])
    index_max_cc = np.zeros(max_cc.shape[0])
    for i in range(max_cc.shape[0]):
        idx = np.where(cc[:,i] == max_cc[i])
        if idx.shape[0] == 1:
            index_max_cc[i] = idx
        else:
            index_max_cc[i] = idx[0]
    
    return index_max_cc


def connectopic_laplacian(data_input_filename, data_roi_filename, mask_filename, data_output_filename, output_eval_filename, output_grad_filename, num_gradients=2, fwhm=None, filtering=False, figures=False, cortical=False):
    """
    Main function to calculate the connectopic Laplacian gradients of an ROI volume in NIFTI format.

    Parameters
    ----------
    data_input_filename : str
        Filename of input volume timeseries
    data_roi_filename : str
        Filename of input volume where the relevant ROI have voxel intensity values = 1
    mask_filename : str
        Filename of input mask of GM where the relevant voxels have intensity = 1
    data_output_filename : str
        Filename of nifti file where the output gradients will be stored
    output_eval_filename : str
        Filename of text file where the output eigenvalues will be stored
    output_grad_filename : str
        Filename of text file where the output gradients will be stored
    num_gradients : int, optional
        Number of non-zero gradients to be calculated. The default is 2.
    figures : bool, optional
        Flag whether to plot gradient figures. The default is True.

    """
    if fwhm:
        print('Smoothing with Gaussian kernel of FWHM = {}mm'.format(fwhm))
        data_input_filename = smooth(data_input_filename, fwhm)
    if filtering is True:
        print("Normalizing input data before filtering")
        img_input = image.load_img(data_input_filename)
        img_mask = image.load_img(mask_filename)
        
        # mask and normalize input data
        data_msk = masking.apply_mask(img_input, img_mask)
        data_msk = normalize_data(data_msk)
        img_input = masking.unmask(data_msk, img_mask)
        
        output_filename = data_input_filename.replace('.nii', '_normalized.nii')
        
        print('Saving normalized data to {}'.format(output_filename))
        nib.save(img_input, output_filename)
        
        print('Performing Wishart filtering using icaDim.m, see: <https://github.com/Washington-University/HCPpipelines/tree/master/global/matlab/icaDim>')
        data_input_filename = wishart(output_filename, mask_filename)
    
    print('Loading images')
    img_input = image.load_img(data_input_filename)
    img_roi = image.load_img(data_roi_filename)
    img_mask = image.load_img(mask_filename)
    
    evals, egrads = calc_gradients(img_input, img_roi, img_mask, num_gradients)
    
    # prepare nifti output array
    roi_data = img_roi.get_fdata()
    inds_all = np.where(roi_data==1)
    xx, yy, zz = inds_all
    
    # initialize nifti output array
    new_shape = np.array(roi_data.shape)
    if roi_data.ndim > 3:
        new_shape[3] = num_gradients
    else:
        new_shape = np.append(new_shape, num_gradients)
    new_data = np.zeros(new_shape)
    
    # put gradients into nifti array
    for grad in range(num_gradients):
        data = egrads[:, grad]
        new_data[xx, yy, zz, grad] = data
        
    # save gradients to nifti
    img = nib.Nifti1Image(new_data, img_roi.affine, header=img_roi.header)
    print('Saving output gradient file to {}'.format(data_output_filename))
    nib.save(img, data_output_filename)
    
    # write eigenvalues and gradient files
    if output_eval_filename:
        print('Saving output eigenvalues file to {}'.format(output_eval_filename))
        np.savetxt(output_eval_filename, evals)
    if output_grad_filename:
        print('Saving output gradients file to {}'.format(output_grad_filename))
        np.savetxt(output_grad_filename, egrads)
        
    if cortical is True:
        print('Computing cortical projection')
        eigvec_proj = cortical_projection(img_input, img_roi, img_mask, egrads)
        cort_img = nib.Nifti1Image(eigvec_proj, img_mask.affine, header=img_mask.header)
        cort_filename = data_output_filename.replace('.nii.gz','_cortical_projection.nii.gz')
        print('Saving cortical projection file to {}'.cort_filename)
        
        nib.save(cort_img, cort_filename)
        
        
    
    # save figures TODO
    


def main(raw_args=None):
    
    #fmt = "%(asctime)s,%(msecs)d %(name)-2s " "%(levelname)-2s:\n\t %(message)s"
    #datefmt = "%y%m%d-%H:%M:%S"
    
    print('Started connectopic Laplacian, parsing command line arguments...')   
    
    parser = ArgumentParser(epilog="connectopic_laplacian.py -- A function to calculate the connectopic Laplacian of an ROI volume. Nikitas C. Koussis, Systems Neuroscience Group, 2023 <nikitas.koussis@gmail.com>")
    parser.add_argument("data_input_filename", help="An input nifti with fmri data", metavar="fmri_input.nii.gz")
    parser.add_argument("data_roi_filename", help="An input nifti with ROI voxel values=1", metavar="roi_input.nii.gz")
    parser.add_argument("mask_filename", help="An input nifti mask with voxel values=1", metavar="gm_mask.nii.gz")
    parser.add_argument("-o", dest="data_output_filename", help="An output nifti where the gradients in volume space will be stored", metavar="gradients.nii.gz")
    parser.add_argument("--eval", dest="output_eval_filename", default=None, help="An output text file where the eigenvalues will be stored", metavar="evals.txt")
    parser.add_argument("--grads", dest="output_grad_filename", default=None, help="An output text file where the gradients will be stored", metavar="grads.txt")
    parser.add_argument("-N", dest="num_gradients", default=2, help="Number of gradients to be calculated, default=2", metavar="2")
    parser.add_argument("--smoothing", dest="smoothing",default=None, help="Option whether to perform smoothing and what FWHM to perform smoothing", metavar="6")
    parser.add_argument("--filter", dest="filtering", action='store_true', default=False, help="Option whether to perform Wishart filtering, INITIALIZES MATLAB SUBROUTINE. REQUIRES MATLAB 2017b OR GREATER")
    parser.add_argument("--output_figures", dest="figures", default=False, help="Option whether to output orthographic figures of gradients")
    parser.add_argument("--cortical", action='store_true', default=False, help="Option whether to output cortical projections of gradients")
    
    #-----------------   Parsing mandatory inputs from terminal:   ----------------
    args = parser.parse_args()
    data_input_filename    = args.data_input_filename
    data_roi_filename      = args.data_roi_filename
    mask_filename          = args.mask_filename
    data_output_filename   = args.data_output_filename
    num_gradients          = int(args.num_gradients)
    output_eval_filename   = args.output_eval_filename
    output_grad_filename   = args.output_grad_filename
    fwhm                   = args.smoothing
    filtering              = args.filtering
    figures                = args.figures
    cortical               = args.cortical
    #-------------------------------------------------------------------------------
    
    print("")
    print("####################### USUAL USAGE ######################")
    print("")
    print("!/bin/bash")
    print('$ data_input_filename=fmri_input.nii.gz \ ')
    print('$ data_roi_filename=roi.nii.gz \ ')
    print('$ mask_filename=mask.nii.gz \ ')
    print('$ data_output_filename=gradients.nii.gz \ ')
    print('$ num_gradients=20 \ ')
    print('$ fwhm=6 \ ')
    print('$ python connectopic_laplacian.py ${data_input_filename} ${data_roi_filename} ${mask_filename} ${data_output_filename} -N ${num_gradients} --smoothing ${fwhm} --filter --cortical')
    print("")
    print('########################### RUN ##########################')
    print("")
    print('Input fmri filename: {}'.format(data_input_filename))
    print('Input ROI filename: {}'.format(data_roi_filename))
    print('Mask filename: {}'.format(mask_filename))
    print('Output filename: {}'.format(data_output_filename))
    # if output_eval_filename:
    #     print('Output eigenvalue filename: {}'.format(output_eval_filename))
    #     if '.txt' not in output_eval_filename:
    #         *_, ext = split_filename(output_eval_filename)
    #         if ext is not None:
    #             print('Output evals file must be saved as .txt file')
    #             print('Giving evals output file .txt extension')
    #             output_eval_filename = output_eval_filename - ext + '.txt'
    #         else:
    #             print('Giving evals output file .txt extension')
    #             output_eval_filename += '.txt'
                
    # if output_grad_filename:
    #     print('Output gradients: {}'.format(output_grad_filename))
    #     if '.txt' not in output_grad_filename:
    #         *_, ext = split_filename(output_grad_filename)
    #         if ext is not None:
    #             print('Output grads file must be saved as .txt file')
    #             print('Giving grads output file .txt extension')
    #             output_grad_filename = output_grad_filename - ext + '.txt'
    #         else:
    #             print('Giving grads output file .txt extension')
    #             output_grad_filename += '.txt'
                
    print('Number of gradients to compute: {}'.format(num_gradients))
    
    if args.smoothing is not None:
        fwhm = float(args.smoothing)
        print('Performing smoothing with {}mm FWHM'.format(fwhm))
    else:
        print('Not performing smoothing')
    
    if filtering is True:
        print('Performing filtering')
        # make sure matlab is installed and on the PATH
    else:
        print('Not performing filtering')
        
    if cortical is True:
        print('Performing cortical projection')
    else:
        print('Not performing cortical projection')
        
    ### TODO FIGURES ###
    
    print('Checking inputs...')
    
    # check input file
    # if not isfile(data_input_filename):
    #     raise OSError('Could not read input file {}'.format(data_input_filename))
    # *_, ext = split_filename(data_input_filename)
    # if ext not in image_types:
    #     raise ValueError('Input file expected to be nifti with extension .nii or .nii.gz, got {}'.format(ext))
    # img_input = image.load_img(data_input_filename)
    # if img_input.ndim != 4:
    #     raise RuntimeError('Input file expected to be volume timeseries with 4 dimensions, got {} dimensions instead'.format(img_input.ndim))
    
    # if not isfile(data_roi_filename):
    #     raise OSError('Could not read ROI file {}'.format(data_roi_filename))
    # *_, ext = split_filename(data_roi_filename)
    # if ext not in image_types:
    #     raise ValueError('ROI file expected to be nifti with extension .nii or .nii.gz, got {}'.format(ext))
    # img_roi = image.load_img(data_roi_filename)
    # if np.max(img_roi.get_fdata()) != 1:
    #     raise RuntimeError('ROI image file must be a binary mask, i.e., have 1 or 0 intensity values')
    
    # if not isfile(mask_filename):
    #     raise OSError('Could not read mask file {}'.format(mask_filename))
    # *_, ext = split_filename(mask_filename)
    # if ext not in image_types:
    #     raise ValueError('Mask file expected to be nifti with extension .nii or .nii.gz, got {}'.format(ext))
    # img_mask = image.load_img(mask_filename)
    # if np.max(img_mask.get_fdata()) != 1:
    #     raise RuntimeError('Mask image file must be a binary mask, i.e., have 1 or 0 intensity values')
    
    # if (img_roi.affine==img_input.affine).all():
    #     raise ValueError("Input ROI must have the same affine transformation as the fmri input file")
    # if (img_roi.affine==img_mask.affine).all():
    #     raise ValueError("Input ROI must have the same affine transformation as the GM mask file")
    # if (img_mask.affine==img_input.affine).all():
    #     raise ValueError("Input mask must have the same affine transformation as the fmri input file")
    
    # if img_roi.shape != img_input.shape[:-1]:
    #     raise ValueError("Input ROI must have the same volume dimensions as the fmri input file")
    # if img_roi.shape != img_mask.shape:
    #     raise ValueError("Input ROI must have the same dimensions as the GM mask file")
    # if img_mask.shape != img_input.shape[:-1]:
    #     raise ValueError("Input mask must have the same volume dimensions as the fmri input file")
    
    # if image_types not in data_output_filename:
    #     *_, ext = split_filename(data_output_filename)
    #     if ext is not None:
    #         print('Data output file must be saved as .nii or .nii.gz file')
    #         print('Giving data output file .nii.gz extension')
    #         data_output_filename = data_output_filename - ext + '.nii.gz'
    #     else:
    #         print('Giving data output file .nii.gz extension')
    #         data_output_filename += '.nii.gz'
    
    # try:
    #     with open(data_output_filename, 'wb') as fh:
    #         fh.write(b'\x1f\x8b')
    #     with open(output_grad_filename, 'w') as fh:
    #         fh.write('')
    #     with open(output_eval_filename, 'w') as fh:
    #         fh.write('')
    # except:
    #     raise OSError('Could not output file into specified format, check write permissions of output folder')
    
    print('Inputs OK')
    
    connectopic_laplacian(data_input_filename, data_roi_filename, mask_filename, data_output_filename, output_eval_filename, output_grad_filename, num_gradients, fwhm, filtering, figures, cortical)
    
if __name__ == '__main__':
    
    #run on command line
    main()
    



from os import path, system
import nibabel as nib

def load_mesh_vertices(vx):
    """
    Loads mesh vertices given by gifti surface filename 'vx'
    
    Returns
    -------
    vertices : (N,3) np.ndarray
    Coordinates of surface vertices
    
    """

    mesh = nib.load(vx)
    vertices = mesh.agg_data('pointset')
    
    return vertices


# -*- coding: utf-8 -*-
"""
Contains functionality for running spatial null models
Adapted from netneurolab/neuromaps (c) 2023

Added functionality for wavestrapping of spatial null models (c) Nikitas C. Koussis 2023
"""

import os
import tempfile
import nibabel as nib
import numpy as np
from scipy import ndimage
from scipy.spatial.distance import cdist
from packaging import version
try:
    import brainsmash
    from brainsmash.mapgen import Base, Sampled
    _brainsmash_avail = True
except ImportError:
    _brainsmash_avail = False
try:
    from brainspace.null_models.moran import MoranRandomization
    _brainspace_avail = True
except ImportError:
    _brainspace_avail = False

from neuromaps.datasets import fetch_atlas
from neuromaps.datasets.atlases import _sanitize_atlas
from neuromaps.images import load_data, load_nifti, PARCIGNORE
from neuromaps.points import get_surface_distance
from neuromaps.transforms import mni152_to_mni152
from neuromaps.nulls.burt import batch_surrogates
from neuromaps.nulls.spins import (gen_spinsamples, get_parcel_centroids,
                                   load_spins, spin_data, spin_parcels)
HEMI = dict(left='L', lh='L', right='R', rh='R')


_nulls_input_docs = dict(
    data_or_none_surface="""\
data : array_like or path_like or giimg_like or tuple
    Input data from which to generate null maps. If None is provided then
    the resampling array will be returned instead. When a parcellation is
    provided, the data must be parcellated and array-like. Otherwise, the data
    must be a surface-based image (giimg_like, e.g. nib.GiftiImage) or a
    path-like object (`str` or `os.PathLike`) pointing to an image file.\
""",
    data_or_none_parcel="""\
data : (N,) array_like
    Input data from which to generate null maps. The data must be
    parcellated and array-like. If None is provided then the resampling array
    will be returned instead.\
""",
    data_parcel="""\
data : (N,) array_like
    Input data from which to generate null maps. The data must be
    parcellated and array-like.\
""",
    data="""\
data : array_like or path_like or niimg_like or giimg_like or tuple
    Input data from which to generate null maps. When a parcellation is
    provided, the data must be parcellated and array-like. Otherwise, the data
    can either be a volumetric image (niimg_like, e.g. nib.Nifti1Image) or a
    surface-based image (giimg_like, e.g. nib.GiftiImage). Alternatively, it
    can be a path-like object (`str` or `os.PathLike`) pointing to an image
    file.\
""",
    atlas_density_surface="""\
atlas : {'fsLR', 'fsaverage', 'civet'}, optional
    Name of surface atlas on which `data` are defined. Default: 'fsaverage'
density : str, optional
    Density of surface mesh on which `data` are defined. Must be
    compatible with specified `atlas`. Default: '10k'\
""",
    atlas_density="""\
atlas : {'fsLR', 'fsaverage', 'civet', 'mni152'}, optional
    Name of atlas on which `data` are defined. Default: 'fsaverage'
density : str, optional
    Density of atlas on which `data` are defined. Must be
    compatible with specified `atlas`. Default: '10k'\
""",
    parcellation="""\
parcellation : tuple-of-str or os.PathLike, optional
    Filepaths to parcellation images ([left, right] hemisphere) mapping `data`
    to atlas specified by `atlas` and `density`. Should only be supplied
    if `data` represents a parcellated null map. Default: None\
""",
    n_perm="""\
n_perm : int, optional
    Number of null maps or permutations to generate. Default: 1000\
""",
    seed="""\
seed : {int, np.random.RandomState instance, None}, optional
    Seed for random number generation. Default: None\
""",
    spins="""\
spins : array_like or str or os.PathLike
    Filepath to or pre-loaded resampling array. If not specified spins are
    generated. Default: None\
""",
    surfaces="""\
surfaces : tuple-of-str or os.PathLike, optional
    Instead of specifying `atlas` and `density` this specifies the surface
    files on which `data` are defined. Providing this will override arguments
    supplied to `atlas` and `density`. Default: None
""",
    n_proc="""\
n_proc : int, optional
    Number of processors to use for parallelizing computations. If negative
    will use max available processors plus 1 minus the specified number.
    Default: 1 (no parallelization)\
""",
    distmat="""\
distmat : tuple-of-str or os.PathLike, optional
    Filepaths to pre-computed (left, right) surface distance matrices.
    Providing this will cause `atlas`, `density`, and `parcellation` to be
    ignored. Default: None\
""",
    kwargs="""\
kwargs : key-value pairs
    Other keyword arguments passed directly to the underlying null method
    generator\
""",
    nulls="""\
nulls : np.ndarray
    Generated null distribution, where each column represents a unique null
    map\
"""
)


def alexander_bloch(data, atlas='fsaverage', density='10k', parcellation=None,
                    n_perm=1000, seed=None, spins=None, surfaces=None):
    if spins is None:
        if surfaces is None:
            surfaces = fetch_atlas(atlas, density)['sphere']
        coords, hemi = get_parcel_centroids(surfaces,
                                            parcellation=parcellation,
                                            method='surface')
        spins = gen_spinsamples(coords, hemi, n_rotate=n_perm, seed=seed)
    spins = load_spins(spins)
    if data is None:
        data = np.arange(len(spins))
    return load_data(data)[spins]


alexander_bloch.__doc__ = """\
Generates null maps from `data` using method from [SN1]_

Method projects data to a spherical surface and uses arbitrary rotations to
generate null distribution. If `data` are parcellated then parcel centroids
are projected to surface and parcels are reassigned based on minimum distances.

Parameters
----------
{data_or_none_surface}
{atlas_density_surface}
{parcellation}
{n_perm}
{seed}
{spins}
{surfaces}

Returns
-------
{nulls}

References
----------
.. [SN1] Alexander-Bloch, A., Shou, H., Liu, S., Satterthwaite, T. D.,
   Glahn, D. C., Shinohara, R. T., Vandekar, S. N., & Raznahan, A. (2018).
   On testing for spatial correspondence between maps of human brain
   structure and function. NeuroImage, 178, 540-51.
""".format(**_nulls_input_docs)


vazquez_rodriguez = alexander_bloch


def vasa(data, atlas='fsaverage', density='10k', parcellation=None,
         n_perm=1000, seed=None, spins=None, surfaces=None):
    if parcellation is None:
        raise ValueError('Cannot use `vasa()` null method without specifying '
                         'a parcellation. Use `alexander_bloch() instead if '
                         'working with unparcellated surface data.')
    if spins is None:
        if surfaces is None:
            surfaces = fetch_atlas(atlas, density)['sphere']
        coords, hemi = get_parcel_centroids(surfaces,
                                            parcellation=parcellation,
                                            method='surface')
        spins = gen_spinsamples(coords, hemi, method='vasa', n_rotate=n_perm,
                                seed=seed)
    spins = load_spins(spins)
    if data is None:
        data = np.arange(len(spins))
    return load_data(data)[spins]


vasa.__doc__ = """\
Generates null maps for parcellated `data` using method from [SN2]_

Method projects parcels to a spherical surface and uses arbitrary rotations
with iterative reassignments to generate null distribution. All nulls are
"perfect" permutations of the input data (at the slight expense of spatial
topology)

Parameters
----------
{data_or_none_parcel}
{atlas_density_surface}
{parcellation}
{n_perm}
{seed}
{spins}
{surfaces}

Returns
-------
{nulls}

References
----------
.. [SN2] Váša, F., Seidlitz, J., Romero-Garcia, R., Whitaker, K. J.,
   Rosenthal, G., Vértes, P. E., ... & Jones, P. B. (2018). Adolescent
   tuning of association cortex in human structural brain networks.
   Cerebral Cortex, 28(1), 281-294.
""".format(**_nulls_input_docs)


def hungarian(data, atlas='fsaverage', density='10k', parcellation=None,
              n_perm=1000, seed=None, spins=None, surfaces=None):
    if parcellation is None:
        raise ValueError('Cannot use `hungarian()` null method without '
                         'specifying a parcellation. Use `alexander_bloch() '
                         'instead if working with unparcellated surface data.')
    if spins is None:
        if surfaces is None:
            surfaces = fetch_atlas(atlas, density)['sphere']
        coords, hemi = get_parcel_centroids(surfaces,
                                            parcellation=parcellation,
                                            method='surface')
        spins = gen_spinsamples(coords, hemi, method='hungarian',
                                n_rotate=n_perm, seed=seed)
    spins = load_spins(spins)
    if data is None:
        data = np.arange(len(spins))
    return load_data(data)[spins]


hungarian.__doc__ = """\
Generates null maps for parcellated `data` using the Hungarian method ([SN3]_)

Method projects parcels to a spherical surface and uses arbitrary rotations
with reassignments based on optimization via the Hungarian method to generate
null distribution. All nulls are "perfect" permutations of the input data (at
the slight expense of spatial topology)

Parameters
----------
{data_or_none_parcel}
{atlas_density_surface}
{parcellation}
{n_perm}
{seed}
{spins}
{surfaces}

Returns
-------
{nulls}

References
----------
.. [SN3] Kuhn, H. W. (1955). The Hungarian method for the assignment problem.
   Naval Research Logistics Quarterly, 2(1‐2), 83-97.
""".format(**_nulls_input_docs)


def baum(data, atlas='fsaverage', density='10k', parcellation=None,
         n_perm=1000, seed=None, spins=None, surfaces=None):
    if parcellation is None:
        raise ValueError('Cannot use `baum()` null method without specifying '
                         'a parcellation. Use `alexander_bloch() instead if '
                         'working with unparcellated surface data.')
    if surfaces is None:
        surfaces = fetch_atlas(atlas, density)['sphere']
    spins = spin_parcels(surfaces, parcellation,
                         n_rotate=n_perm, spins=spins, seed=seed)
    if data is None:
        data = np.arange(len(spins))
    data = load_data(data)
    nulls = data[spins]
    nulls[spins == -1] = np.nan
    return nulls


baum.__doc__ = """\
Generates null maps for parcellated `data` using method from [SN4]_

Method projects `data` to spherical surface and uses arbitrary rotations to
generate null distributions. Reassigned parcels are based on the most common
(i.e., modal) value of the vertices in each parcel within the the rotated data

Parameters
----------
{data_or_none_parcel}
{atlas_density_surface}
{parcellation}
{n_perm}
{seed}
{spins}
{surfaces}

Returns
-------
{nulls}

References
----------
.. [SN4] Baum, G. L., Cui, Z., Roalf, D. R., Ciric, R., Betzel, R. F., Larsen,
   B., ... & Satterthwaite, T. D. (2020). Development of structure–function
   coupling in human brain networks during youth. Proceedings of the National
   Academy of Sciences, 117(1), 771-778.
""".format(**_nulls_input_docs)


def cornblath(data, atlas='fsaverage', density='10k', parcellation=None,
              n_perm=1000, seed=None, spins=None, surfaces=None):
    if parcellation is None:
        raise ValueError('Cannot use `cornblath()` null method without '
                         'specifying a parcellation. Use `alexander_bloch() '
                         'instead if working with unparcellated surface data.')
    data = load_data(data)
    if surfaces is None:
        surfaces = fetch_atlas(atlas, density)['sphere']
    nulls = spin_data(data, surfaces, parcellation,
                      n_rotate=n_perm, spins=spins, seed=seed)
    return nulls


cornblath.__doc__ = """\
Generates null maps for parcellated `data` using method from [SN5]_

Method projects `data` to spherical surface and uses arbitrary rotations to
generate null distributions. Reassigned parcels are based on the average value
of the vertices in each parcel within the rotated data

Parameters
----------
{data_parcel}
{atlas_density_surface}
{parcellation}
{n_perm}
{seed}
{spins}
{surfaces}

Returns
-------
{nulls}

References
----------
.. [SN5] Cornblath, E. J., Ashourvan, A., Kim, J. Z., Betzel, R. F., Ciric, R.,
   Adebimpe, A., ... & Bassett, D. S. (2020). Temporal sequences of brain
   activity at rest are constrained by white matter structure and modulated by
   cognitive demands. Communications biology, 3(1), 1-12.
""".format(**_nulls_input_docs)


def _get_distmat(hemisphere, atlas='fsaverage', density='10k',
                 parcellation=None, drop=None, n_proc=1):
    hemi = HEMI.get(hemisphere, hemisphere)
    if hemi not in ('L', 'R'):
        raise ValueError(f'Invalid hemishere designation {hemisphere}')

    if drop is None:
        drop = PARCIGNORE

    atlas = fetch_atlas(atlas, density)
    surf, medial = getattr(atlas['pial'], hemi), getattr(atlas['medial'], hemi)
    if parcellation is None:
        dist = get_surface_distance(surf, medial=medial, n_proc=n_proc)
    else:
        dist = get_surface_distance(surf, parcellation=parcellation,
                                    medial_labels=drop, drop=drop,
                                    n_proc=n_proc)
    return dist


_get_distmat.__doc__ = """\
Generates surface distance matrix for specified `hemisphere`

If `parcellation` is provided then the returned distance matrix will be a
parcel-parcel matrix.

Parameters
----------
hemisphere : {{'L', 'R'}}
    Hemisphere of surface from which to generate distance matrix
{atlas_density_surface}
{parcellation}
drop : list-of-str, optional
    If `parcellation` is not None, which parcels should be ignored / dropped
    from the generate distance matrix. If not specified will ignore parcels
    generally indicative of the medial wall. Default: None
{n_proc}

Returns
-------
dist : (N, N) np.ndarray
    Surface distance matrix between vertices. If a `parcellation` is specified
    then this will be the parcel-parcel distance matrix, where the distance
    between parcels is the average distance between all constituent vertices
""".format(**_nulls_input_docs)


def _surf_surrogates(data, atlas, density, parcellation, distmat, n_proc):

    data = load_data(data)

    if parcellation is None:
        parcellation = (None, None)

    for n, (hemi, parc) in enumerate(zip(('L', 'R'), parcellation)):
        if distmat is None:
            dist = _get_distmat(hemi, atlas=atlas, density=density,
                                parcellation=parc, n_proc=n_proc)
        else:
            dist = distmat[n]

        if parc is None:
            idx = np.arange(n * (len(data) // 2), (n + 1) * (len(data) // 2))
        else:
            idx = np.trim_zeros(np.unique(load_data(parc))) - 1

        hdata = np.squeeze(data[idx])
        med = np.isinf(dist + np.diag([np.inf] * len(dist))).all(axis=1)
        mask = np.logical_not(np.logical_or(np.isnan(hdata), med))

        yield hdata[mask], dist[np.ix_(mask, mask)], None, idx[mask]


def _vol_surrogates(data, atlas, density, parcellation, distmat, **kwargs):

    if atlas != 'MNI152':
        raise ValueError('Cannot compute volumetric surrogates if atlas is '
                         f'not "MNI152". Received: {atlas}')

    if parcellation is not None:
        try:
            data = np.asarray(data, dtype='float32').squeeze()
        except (TypeError, ValueError) as err:
            msg = ('`data` must be array_like and represent a '
                   'parcellated brain map when `parcellation` is not '
                   '"None"')
            raise TypeError(msg) from err
        if data.ndim != 1:
            raise ValueError('Brain map must be one-dimensional. Got shape '
                             f'{data.shape}')

    atlas = fetch_atlas(atlas, density)

    # get data + coordinates of valid datapoints
    if parcellation is None:
        darr = load_data(data)
        if nib.load(atlas['2009cAsym_T1w']).shape == darr.shape:
            bmask = atlas['2009cAsym_brainmask']
        elif (density in ('1mm', '2mm')
              and nib.load(atlas['6Asym_T1w']).shape == darr.shape):
            bmask = atlas['6Asym_brainmask']
        else:
            bmask = mni152_to_mni152(nib.load(atlas['2009cAsym_brainmask']),
                                     data,
                                     'nearest')
        darr *= load_data(bmask)
        affine = load_nifti(bmask).affine
    else:
        darr = load_data(parcellation)
        affine = load_nifti(parcellation).affine
        labels = np.trim_zeros(np.unique(darr))
    mask = np.logical_not(np.logical_or(np.isclose(darr, 0), np.isnan(darr)))
    xyz = nib.affines.apply_affine(affine, np.column_stack(np.where(mask)))

    # calculate distance matrix
    index = None
    if distmat is None:
        if parcellation is None:  # memmap because BIG
            distout = tempfile.NamedTemporaryFile(suffix='.mmap')
            dist = np.memmap(distout, mode='w+', dtype='float32',
                             shape=(len(xyz), len(xyz)))
            indout = tempfile.NamedTemporaryFile(suffix='.mmap')
            index = np.memmap(indout, mode='w+', dtype='int32',
                              shape=(len(xyz), len(xyz)))
            for n, row in enumerate(xyz):
                xyz_dist = cdist(row[None], xyz).astype('float32')
                index[n] = np.argsort(xyz_dist, axis=-1).astype('int32')
                dist[n] = xyz_dist[:, index[n]]
                dist.flush()
                index.flush()
        else:
            parcellation = darr[mask]
            row_dist = np.zeros((len(xyz), len(labels)), dtype='float32')
            dist = np.zeros((len(labels), len(labels)), dtype='float32')
            for n, row in enumerate(xyz):
                xyz_dist = cdist(row[None], xyz).astype('float32')
                row_dist[n] = ndimage.mean(xyz_dist, parcellation, labels)
            for n in range(len(labels)):
                dist[n] = ndimage.mean(row_dist[:, n], parcellation, labels)

    else:
        dist = distmat
        if parcellation is None:
            index = np.argsort(dist, axis=-1)

    if parcellation is None:
        yield darr[mask], dist, index, mask
    else:
        yield data, dist, index, np.ones(len(labels), dtype=bool)


def _make_surrogates(data, method, atlas='fsaverage', density='10k',
                     parcellation=None, n_perm=1000, seed=None, distmat=None,
                     n_proc=1, **kwargs):
    if method not in ('burt2018', 'burt2020', 'moran'):
        raise ValueError(f'Invalid null method: {method}')

    atlas = _sanitize_atlas(atlas)
    darr = load_data(data)
    surrogates = np.full((darr.shape) + (n_perm,), np.nan)
    genfunc = _vol_surrogates if atlas == 'MNI152' else _surf_surrogates
    for hdata, hdist, hind, hsl in genfunc(data, atlas, density,
                                           parcellation, distmat,
                                           n_proc=n_proc):
        if method == 'burt2018':
            hdata += np.abs(np.nanmin(darr)) + 0.1
            hsurr = batch_surrogates(hdist, hdata, n_surr=n_perm, seed=seed)
        elif method == 'burt2020':
            if parcellation is None:
                if hind is None:
                    hind = np.argsort(hdist, axis=-1)
                    hdist = np.sort(hdist, axis=-1)
                hsurr = Sampled(hdata, hdist, hind, n_jobs=n_proc,
                                seed=seed, **kwargs)(n_perm).T
            else:
                hsurr = Base(hdata, hdist, seed=seed, **kwargs)(n_perm, 50).T
            if hsurr.ndim == 1:
                hsurr = np.expand_dims(hsurr, axis=1)
            # clean up (FIXME: this is ugly and we should fix it)
            if hasattr(hdist, 'filename'):
                hdist._mmap.close()
                hind._mmap.close()
                os.unlink(hdist.filename)
                os.unlink(hind.filename)
                del hdist, hind
        elif method == 'moran':
            dist = hdist.astype('float64')
            np.fill_diagonal(dist, 1)
            dist **= -1
            opts = dict(joint=True, tol=1e-6, n_rep=n_perm, random_state=seed)
            opts.update(**kwargs)
            mrs = MoranRandomization(**kwargs)
            hsurr = mrs.fit(dist).randomize(hdata).T

        surrogates[hsl] = hsurr

    return surrogates


_make_surrogates.__doc__ = """\
Generates null surrogates for specified `data` using `method`

Parameters
----------
{data}
method : {{'burt2018', 'burt2020', 'moran'}}
    Method by which to generate null surrogates
{atlas_density}
{parcellation}
{n_perm}
{seed}
{distmat}
{n_proc}
{kwargs}

Returns
-------
{nulls}
""".format(**_nulls_input_docs)


def burt2018(data, atlas='fsaverage', density='10k', parcellation=None,
             n_perm=1000, seed=None, distmat=None, n_proc=1, **kwargs):
    return _make_surrogates(data, 'burt2018', atlas=atlas, density=density,
                            parcellation=parcellation, n_perm=n_perm,
                            seed=seed, n_proc=n_proc, distmat=distmat,
                            **kwargs)


burt2018.__doc__ = """\
Generates null maps for `data` using method from [SN6]_

Method uses a spatial auto-regressive model to estimate distance-dependent
relationship of `data` and generates surrogate maps with similar properties

Parameters
----------
{data}
{atlas_density}
{parcellation}
{n_perm}
{seed}
{distmat}
{kwargs}

Returns
-------
{nulls}

References
----------
.. [SN6] Burt, J. B., Demirtaş, M., Eckner, W. J., Navejar, N. M., Ji, J. L.,
   Martin, W. J., ... & Murray, J. D. (2018). Hierarchy of transcriptomic
   specialization across human cortex captured by structural neuroimaging
   topography. Nature Neuroscience, 21(9), 1251-1259.
""".format(**_nulls_input_docs)


def burt2020(data, atlas='fsaverage', density='10k', parcellation=None,
             n_perm=1000, seed=None, distmat=None, n_proc=1, **kwargs):
    if not _brainsmash_avail:
        raise ImportError('Cannot run burt2020 null model when `brainsmash` '
                          'is not installed. Please `pip install brainsmash` '
                          'and try again.')
    elif version.parse(brainsmash.__version__) < version.parse('0.10.0'):
        raise ImportError('The burt2020 null model needs version >= 0.10.0 of '
                          '`brainsmash. Please `pip install brainsmash`'
                          '`--upgrade` and try again.')
    return _make_surrogates(data, 'burt2020', atlas=atlas, density=density,
                            parcellation=parcellation, n_perm=n_perm,
                            seed=seed, n_proc=n_proc, distmat=distmat,
                            **kwargs)


burt2020.__doc__ = """\
Generates null maps for `data` using method from [SN7]_ and [SN8]_

Method uses variograms to estimate spatial autocorrelation of `data` and
generates surrogate maps with similar variogram properties

Parameters
----------
{data}
{atlas_density}
{parcellation}
{n_perm}
{seed}
{n_proc}
{distmat}
{kwargs}

Returns
-------
{nulls}

References
----------
.. [SN7] Burt, J. B., Helmer, M., Shinn, M., Anticevic, A., & Murray, J. D.
   (2020). Generative modeling of brain maps with spatial autocorrelation.
   NeuroImage, 220, 117038.
.. [SN8] https://github.com/murraylab/brainsmash
""".format(**_nulls_input_docs)


def moran(data, atlas='fsaverage', density='10k', parcellation=None,
          n_perm=1000, seed=None, distmat=None, n_proc=1, **kwargs):
    if not _brainspace_avail:
        raise ImportError('Cannot run moran null model when `brainspace` is '
                          'not installed. Please `pip install brainspace` and '
                          'try again.')
    return _make_surrogates(data, 'moran', atlas=atlas, density=density,
                            parcellation=parcellation, n_perm=n_perm,
                            seed=seed, n_proc=n_proc, distmat=distmat,
                            **kwargs)


moran.__doc__ = """\
Generates null maps for `data` using method from [SN9]_

Method uses a spatial decomposition of a distance-based weight matrix to
estimate eigenvectors that are used to generate surrogate maps by imposing a
similar spatial structure on randomized data. For a MATLAB implementation
refer to [SN10]_ and [SN11]_

Parameters
----------
{data}
{atlas_density}
{parcellation}
{n_perm}
{seed}
{n_proc}
{distmat}
{kwargs}

Returns
-------
{nulls}

References
----------
.. [SN9] Wagner, H. H., & Dray, S. (2015). Generating spatially constrained
   null models for irregularly spaced data using M oran spectral randomization
   methods. Methods in Ecology and Evolution, 6(10), 1169-1178.
.. [SN10] de Wael, R. V., Benkarim, O., Paquola, C., Lariviere, S., Royer, J.,
   Tavakol, S., ... & Bernhardt, B. C. (2020). BrainSpace: a toolbox for the
   analysis of macroscale gradients in neuroimaging and connectomics datasets.
   Communications Biology, 3(1), 1-10.
.. [SN11] https://github.com/MICA-MNI/BrainSpace/
""".format(**_nulls_input_docs)



import numpy as np
from joblib import Parallel, delayed
from numpy.random import Generator, PCG64

class Permutation:
    """
    Parallel implementation of permutation of `np.ndarray`, where the objective 
    is to calculate a null distribution of group average metric files '.func.gii' 
    for significance testing.
            
    Parameters
    ----------
    x : (n,N) np.ndarray
        Array of n subjects by N = `n_vertices`. NOTE: Must be in common space
        {'fsLR', 'fsaverage'}. See neuromaps.datasets.fetch_atlas()
        
    y : (m,N) np.ndarray
        Array of m subjects by N = `n_vertices`. NOTE: Must be in common space
        {'fsLR', 'fsaverage'}. See neuromaps.datasets.fetch_atlas()
        
        NOTE
        ----
        n does not have to equal m but N must be the same.
    
    n_perm : int
        Number of permutations
        
    seed : None or int
        Specify the seed for random number generation
        
    n_jobs : int
        Number of jobs to use for parallelization
        
    Returns
    -------
    (n_perm,n,N) np.ndarray
        All surrogates for group x
    
    (n_perm,m,N) np.ndarray
        All surrogates for group y
    
    Raises
    ------
    ValueError: `x` or `y` is not np.ndarray
    RuntimeError: `x` and `y` do not have the same number of vertices
    
    """
    def __init__(self, x, y, n_perm=100, seed=None, n_jobs=1):
        
        self._n_perm = n_perm
        self._seed = seed
        self._n_jobs = n_jobs
        
        if type(x) is not np.ndarray or type(y) is not np.ndarray:
            raise ValueError("Input is not `np.ndarray`, is {} instead".format(type(x)))
        if x.shape[1] != y.shape[1]:
            raise RuntimeError("Inputs `x` and `y` must have the same number of vertices")
        
        self._x = x
        self._y = y
        
        if seed == None:
            self._rng = np.random.default_rng()
        else:
            self._rng = Generator(PCG64(seed))
        
    def __call__(self, n_perm):
        """
        Generate permuted arrays for arrays given by `x` and `y`

        Parameters
        ----------
        n_perm : number of permuted arrays to return

        Returns
        -------
        (n,N) np.ndarray
            Shuffled group x
        (n,N) np.ndarray
            Shuffed group y

        """
        rs = np.arange(n_perm)
        surr_x, surr_y = np.stack(
            Parallel(
                self._n_jobs, prefer='threads')(
                delayed(self._call_method)(self._x, self._y, self._rng, rs=i) for i in rs
            ), axis=1
        )
        return [np.asarray([*surr_x]).squeeze(), 
                np.asarray([*surr_y]).squeeze()
            ]
                        
        
    def _call_method(self, x, y, rng, rs):
        """
        Subfunction used by .permutation() for parallelization purposes
        """    
        #generate new arrays        
        surr_x, surr_y = self.permute_array(x, y, rng)
        
        return surr_x.squeeze(), surr_y.squeeze()
            
    
    def permute_array(self, x, y, rng):
        """
        Generates permuted subject lists from `x` to generate null distribution.
    
        Parameters
        ----------
        x : (n,N) np.ndarray
        
        y : (m,N) np.ndarray
        
            NOTE
            ----
            n does not have to equal m but N must be the same.
            
        rng : None or np.random.Generator method
            Set for reproducibility, default None
    
        Returns
        -------
        surr_x : (n,N) np.ndarray
            Surrogate array from reshuffling random m subjects into n
            
        surr_y : (m,N) np.ndarray
            Surrogate array from reshuffling random n subjects into m
    
        """
        n = x.shape[0]
        #m = y.shape[0]
        
        #pool groups to shuffle
        data = np.vstack((x, y))
        #shuffle
        rng.shuffle(data, axis=0)
        
        surr_x, surr_y = data[:n], data[n+1:]
        
        return surr_x, surr_y
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Eigenmode resampling on the cortex
"""

from neuroshape.utils.eigen import (
    _get_eigengroups,
    reconstruct_surface,
    transform_to_spheroid,
    transform_to_ellipsoid,
    resample_spheroid,
    )

import numpy as np
import nibabel as nib
import matplotlib.pyplot as plt
from nilearn import plotting

cmap = plt.get_cmap('viridis')

norm_types = [
    'constant',
    'number',
    'volume',
    'area',
    ]

methods = [
    'matrix',
    'matrix_separate',
    'regression',
    ]

def eigenmode_resample(surface, evals, emodes, angles=None, normalize='area', 
                         decomp_method='matrix', norm_factor=1):
    """
    Resample the hypersphere bounded by the eigengroups contained within `emodes`.
    Based on the degenerate solutions of solving the Laplace-Beltrami operator 
    on the cortex. The power spectrum is perfectly retained (the square of the 
    eigenvalues).
    
    @author: Nikitas C. Koussis, School of Psychological Sciences,
             University of Newcastle & Systems Neuroscience Group Newcastle
             
             Michael Breakspear, School of Psychological Sciences,
             University of Newcastle & Systems Neuroscience Group Newcastle
    
    Process
    -------
        - the orthonormal eigenvectors n within eigengroup l give the surface 
            of the hyperellipse of dimensions l-1
            NOTE: for eigengroups 1, this is resampling the line,
            and for eigengroup 2, this is resampling the surface of the 
            2-sphere
        - the axes of the hyperellipse are given by the sqrt of the eigenvalues 
            corresponding to eigenvectors n
        - linear transform the eigenvectors N to the hypersphere by dividing 
            by the ellipsoid axes
        - find the set of points `p` on the hypersphere given by the basis 
            modes (the eigenmodes)
        - rotate the set of points `p` by a set given by `angles` of (0, 2*pi) for 
            even dimensions and (0, pi) for odd dimensions (resampling step)
        - find the unit vectors of the points `p` by dividing by the 
            Euclidean norm (equivalent to the eigenmodes)
        - make the new unit vectors orthonormal using QR decomposition
        - return the new eigenmodes of that eigengroup until all eigengroups
            are computed
        - reconstruct the new surface using the new eigenmodes and normalize
            the new vertices to avoid weirdness and discontinuities
    
    References
    ----------
        1. Robinson, P. A. et al., (2016), Eigenmodes of brain activity: Neural field 
        theory predictions and comparison with experiment. NeuroImage 142, 79-98.
        <https://dx.doi.org/10.1016/j.neuroimage.2016.04.050>
        
        2. Jorgensen, M., (2014), Volumes of n-dimensional spheres and ellipsoids. 
        <https://www.whitman.edu/documents/Academics/Mathematics/2014/jorgenmd.pdf>
        
        3. <https://math.stackexchange.com/questions/3316373/ellipsoid-in-n-dimension>
        
        4. Blumenson, L. E. (1960). A derivation of n-dimensional spherical 
        coordinates. The American Mathematical Monthly, 67(1), 63-66.
        <https://www.jstor.org/stable/2308932>
        
        5. Trefethen, Lloyd N., Bau, David III, (1997). Numerical linear algebra. 
        Philadelphia, PA: Society for Industrial and Applied Mathematics. 
        ISBN 978-0-898713-61-9.
        
        6. <https://en.wikipedia.org/wiki/QR_decomposition>
        
        7. Chen, Y. C. et al., (2022). The individuality of shape asymmetries 
        of the human cerebral cortex. Elife, 11, e75056. 
        <https://doi.org/10.7554/eLife.75056>
        

    Parameters
    ----------
    surface : nib.GiftiImage class
        Surface to resample of surface.darrays[0].data of shape (n_vertices, 3)
        and surface.darrays[1].data of shape (n_faces, 3). Must be a single
        hemisphere or bounded surface.
        
    evals : np.ndarray of shape (n,)
        Eigenvalues corresponding to the number of eigenmodes n in `emodes`
        
    emodes : np.ndarray of shape (n_vertices, n)
        Eigenmodes that are the solution to the generalized eigenvalue problem
        or Helmholtz equation in the Laplace-Beltrami operator of the cortex
        
    angles : np.ndarray of shape (n,), optional
        Angles to pass to eigen.resample_spheroid(). The default is None.
        If none, random angles in the half-open interval between [0, 2pi) are
        passed.
        
    normalize : str, optional
        Normalization method for the vertices of the new surface.
        
        Accepted types:
            'constant' : normalize by a constant factor (default is 1^(1/3))
            'number' : normalize by the number of vertices
            'volume' : normalize by the volume of the reconstructed surface
            'area' : normalize by the area of the new vertices
        
        The default is 'area'.
        
    decomp_method : str, optional
        method of calculation of coefficients: 'matrix', 'matrix_separate', 
        'regression'.
        
        The default is 'matrix'.
        
    norm_factor : int, optional
        Normalization factor for 'constant'. Unused in any other method.

    Returns
    -------
    new_surface : nib.GiftiImage class
        The new surface that has been reconstructed using the new eigenmodes.
        
    Raises
    ------
    ValueError : Inappropriate inputs

    """
    # perform checks
    if emodes.shape[0] != surface.darrays[0].data.shape[0]: 
        # try transpose
        emodes = emodes.T
        if emodes.shape[0] != surface.darrays[0].data.shape[0]:
            raise ValueError("Eigenmodes must have the same number of vertices as the surface")
    if evals.ndim != 1:
        raise ValueError("Eigenvalue array must be 1-dimensional")
    if emodes.shape[1] != evals.shape:
        # try transpose
        emodes = emodes.T
        if emodes.shape[1] != evals.shape:
            raise ValueError("There must be as many eigenmodes as there are eigenvalues")
    if not isinstance(surface, nib.GiftiImage):
        raise ValueError("Input surface must be of nibabel.GiftiImage class")
    if emodes.shape[1] >= surface.darrays[0].data.shape[0]:
        raise ValueError("Number of eigenmodes must be less than the number of vertices on the surface")
    if decomp_method in methods:
        method = decomp_method
    else:
        raise ValueError("Eigenmode decomposition method must be 'matrix', 'matrix_separate', 'regression'")
        
    # if not given angles
    if angles is None:
        angles = np.random.random_sample(size=emodes.shape[1])
    
    # find eigengroups
    groups = _get_eigengroups(emodes)
    
    # ensure the eigenmodes are orthonormal
    emodes = emodes / np.linalg.norm(emodes, axis=0)
    
    # initialize the new modes
    new_modes = np.zeros_like(emodes)
    
    # resample the hypersphere (except for groups 1 and 2)
    for group in groups:
        group_modes = emodes[:, group]
        group_evals = evals[group]
        group_new_modes = new_modes[:, group]
        
        if len(group) == 1:
            # resample along the line of real numbers (0, 1)
            group_modes *= np.random.random(0, 1)
            # ensure orthonormal
            new_modes[:, group] = group_modes / np.linalg.norm(group_modes)
        
        if len(group) == 2:
            r_i = 1
            # do simple rotation
            # initialize the points
            p = r_i * group_modes
            for i in range(1, group_modes.shape[1]):
                for j in range(i):
                    r_i *= np.sin(angles[j])
                    if i < group_modes.shape[0] - 1:
                        r_i *= np.cos(angles[j-1])
            
                p += r_i * group_modes[i]
            
            # ensure orthonormal
            group_new_modes = p / np.linalg.norm(p)
            new_modes[:, group], _ = np.linalg.qr(group_new_modes, mode='reduced')
            
        # else, transform to spheroid and index the angles properly
        group_modes = transform_to_spheroid(group_evals, group_modes)
        group_new_modes = resample_spheroid(group_modes, angles[group])
        
        # transform back to ellipsoid
        new_modes[:, group] = transform_to_ellipsoid(group_evals, group_new_modes)
    
    # reconstruct the new surface
    if normalize in norm_types:
        if normalize == 'constant':
            if norm_factor > 0.:
                new_surface = reconstruct_surface(surface, new_modes, normalize=normalize, norm_factor=norm_factor, method=method)
            else:
                raise ValueError("Normalization factor must be greater than zero")
        else:
            new_surface = reconstruct_surface(surface, new_modes, normalize=normalize, norm_factor=norm_factor, method=method)
            
    else:
        raise ValueError("Normalization type must be 'constant', 'number', 'volume', 'area'")
        
    return new_surface


def plot_surface(surface, data=None, hemi='left', view='lateral', cmap='gray', show=True):
    """
    Plots a surface using nilearn.plotting, returns fig and ax handles
    from matplotlib.pyplot for further use. Can also plot values on the
    surface by input to `data`.

    Parameters
    ----------
    surface : nib.GiftiImage class or np.ndarray of shape (n_vertices, 3)
        A single surface to plot.
    data : np.ndarray of shape (n_vertices,)
        Data to plot on the surface
    hemi : str, optional
        Which hemisphere to plot. The default is 'left'.
    view : str, optional
        Which view to look at the surface. 
        The default is 'lateral'. Accepted strings are detailed in
        the docs for nilearn.plotting
    cmap : str or matplotlib.cm class
        Which colormap to plot the surface with, default is 'viridis'
    show : bool, optional
        Flag whether to show the plot, default is True

    Returns
    -------
    fig : figure handle    
    ax : axes handle

    """
    # make figure
    fig = plt.figure(figsize=(15,9), constrained_layout=False)
    mesh = surface.darrays
    
    # get colormap
    cmap = plt.get_cmap(cmap)
    
    # check if data
    if data is not None:
        if data.shape != surface.darrays[0].data.shape[0]:
            raise ValueError("Number of data points must be the same as the number of surface vertices")
        
        vmin = np.min(data)
        vmax = np.max(data)
        
    else:
        vmin = None
        vmax = None
    
    # plot surface
    ax = fig.add_subplot(projection='3d')
    plotting.plot_surf(mesh, surf_map=data, hemi=hemi, view=view, 
                       vmin=vmin, vmax=vmax, colorbar=False, 
                       cmap=cmap, axes=ax)
    ax.dist = 7
    
    # show figure check
    if show is True:
        plt.show()
    
    return fig, ax
    

def eigen_spectrum(evals, show=True):
    """
    Plot the eigenvalue power spectrum. Returns fig and ax handles from
    matplotlib.pyplot for further use.

    Parameters
    ----------
    evals : np.ndarray
        Number of eigenvalues
    show : bool, optional
        Flag whether to show plot, default True

    Returns
    -------
    fig : matplotlib.pyplot class
        Figure handle
    ax : matplotlib.pyplot class
        Axes handle

    """
    
    # compute power spectrum = eval^2
    
    power = evals*evals
    
    # now do figure
    fig = plt.figure(figsize=(15, 9), constrained_layout=False)
    
    ax = fig.add_subplot()
    
    markerline, _, _ = plt.stem(np.arange(1,len(evals)+1), power, linefmt='blue', markerfmt=None)
    
    plt.xlabel(r'Eigenvalue $\lambda$')
    plt.ylabel(r'Eigenvalue power $\lambda^2$')
    
    # show figure check
    if show is True:
        plt.show()
    
    return fig, ax


'''
Copyright (C) 2018  Andre Lester Kruger
ConcaveHull.py is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.
ConcaveHull.py is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with ConcaveHull.py.  If not, see <http://www.gnu.org/licenses/>.
'''

import bisect
from collections import OrderedDict
import math
#import numpy as np
import matplotlib.tri as tri
from shapely.geometry import LineString
from shapely.geometry import Polygon
from shapely.ops import linemerge


class ConcaveHull:
    
    def __init__(self):
        self.triangles = {}
        self.crs = {}
        
    
    def loadpoints(self, points):
        #self.points = np.array(points)
        self.points = points
        
        
    def edge(self, key, triangle):
        '''Calculate the length of the triangle's outside edge
        and returns the [length, key]'''
        pos = triangle[1].index(-1)
        if pos==0:
            x1, y1 = self.points[triangle[0][0]]
            x2, y2 = self.points[triangle[0][1]]
        elif pos==1:
            x1, y1 = self.points[triangle[0][1]]
            x2, y2 = self.points[triangle[0][2]]
        elif pos==2:
            x1, y1 = self.points[triangle[0][0]]
            x2, y2 = self.points[triangle[0][2]]
        length = ((x1-x2)**2+(y1-y2)**2)**0.5
        rec = [length, key]
        return rec
        
    
    def triangulate(self):
        
        if len(self.points) < 2:
            raise Exception('CountError: You need at least 3 points to Triangulate')
        
        temp = list(zip(*self.points))
        x, y = list(temp[0]), list(temp[1])
        del(temp)
        
        triang = tri.Triangulation(x, y)
        
        self.triangles = {}
        
        for i, triangle in enumerate(triang.triangles):
            self.triangles[i] = [list(triangle), list(triang.neighbors[i])]
        

    def calculatehull(self, tol=50):
        
        self.tol = tol
        
        if len(self.triangles) == 0:
            self.triangulate()
        
        # All triangles with one boundary longer than the tolerance (self.tol)
        # is added to a sorted deletion list.
        # The list is kept sorted from according to the boundary edge's length
        # using bisect        
        deletion = []    
        self.boundary_vertices = set()
        for i, triangle in self.triangles.items():
            if -1 in triangle[1]:
                for pos, neigh in enumerate(triangle[1]):
                    if neigh == -1:
                        if pos == 0:
                            self.boundary_vertices.add(triangle[0][0])
                            self.boundary_vertices.add(triangle[0][1])
                        elif pos == 1:
                            self.boundary_vertices.add(triangle[0][1])
                            self.boundary_vertices.add(triangle[0][2])
                        elif pos == 2:
                            self.boundary_vertices.add(triangle[0][0])
                            self.boundary_vertices.add(triangle[0][2])
            if -1 in triangle[1] and triangle[1].count(-1) == 1:
                rec = self.edge(i, triangle)
                if rec[0] > self.tol and triangle[1].count(-1) == 1:
                    bisect.insort(deletion, rec)
                    
        while len(deletion) != 0:
            # The triangles with the longest boundary edges will be 
            # deleted first
            item = deletion.pop()
            ref = item[1]
            flag = 0
            
            # Triangle will not be deleted if it already has two boundary edges            
            if self.triangles[ref][1].count(-1) > 1:
                continue
                
            # Triangle will not be deleted if the inside node which is not
            # on this triangle's boundary is already on the boundary of 
            # another triangle
            adjust = {0: 2, 1: 0, 2: 1}            
            for i, neigh in enumerate(self.triangles[ref][1]):
                j = adjust[i]
                if neigh == -1 and self.triangles[ref][0][j] in self.boundary_vertices:
                    flag = 1
                    break
            if flag == 1:
                continue
           
            for i, neigh in enumerate(self.triangles[ref][1]):
                if neigh == -1:
                    continue
                pos = self.triangles[neigh][1].index(ref)
                self.triangles[neigh][1][pos] = -1
                rec = self.edge(neigh, self.triangles[neigh])
                if rec[0] > self.tol and self.triangles[rec[1]][1].count(-1) == 1:
                    bisect.insort(deletion, rec)
                    
            for pt in self.triangles[ref][0]:
                self.boundary_vertices.add(pt)
                                        
            del self.triangles[ref]
            
        self.polygon()
            
                    

    def polygon(self):
        
        edgelines = []
        for i, triangle in self.triangles.items():
            if -1 in triangle[1]:
                for pos, value in enumerate(triangle[1]):
                    if value == -1:
                        if pos==0:
                            x1, y1 = self.points[triangle[0][0]]
                            x2, y2 = self.points[triangle[0][1]]
                        elif pos==1:
                            x1, y1 = self.points[triangle[0][1]]
                            x2, y2 = self.points[triangle[0][2]]
                        elif pos==2:
                            x1, y1 = self.points[triangle[0][0]]
                            x2, y2 = self.points[triangle[0][2]]
                        line = LineString([(x1, y1), (x2, y2)])
                        edgelines.append(line)

        bound = linemerge(edgelines)
    
        self.boundary = Polygon(bound.coords)


from joblib import Parallel, delayed
from neuromaps.stats import compare_images
import numpy as np

def compare_geomodes_parallel(subjects_array, num_modes=200, n_jobs=1):
    num_subjects = subjects_array.shape[0]
    
    jobs = []
    
    for subi in range(num_subjects):
        subject1 = subjects_array[subi]
        
        for subj in range(subi+1, num_subjects):
            subject2 = subjects_array[subj]
            
            jobs.append(delayed(compare_modes)(
                subject1, subject2, num_modes)
                )
            
    corr = np.vstack(
        Parallel(n_jobs=n_jobs, prefer='threads')(jobs)
        )
    
    return np.asarray(corr).squeeze()

def compare_modes(subject1, subject2, num_modes):
    corr = np.zeros(num_modes)
    for j in range(num_modes):
        corr[j] = compare_images(subject1[j], subject2[j], metric='pearsonr')

    return np.asarray(corr).squeeze()


"""
Creates surrogate data by permuting wavelet coefficients for 2-D data
Only shuffles wavelet coefficients where the data is non-zero
Based on M. Breakspear 2002,2022
Updated for Python by N. Koussis 2022
Based on Sampled from Murray lab brainsmash 2022
"""

from ..utils.dataio import dataio
from ..utils.concavehull import ConcaveHull
from ..utils.load_mesh_vertices import load_mesh_vertices
from ..utils.checks import check_map, check_pv, check_deltas
from ..utils.meshtransforms import grid, mesh
from ..utils.shuffle_matrix import shuffle_matrix
from sklearn.linear_model import LinearRegression
from sklearn.utils.validation import check_random_state
import numpy as np
from joblib import Parallel, delayed
import pywt
import nibabel as nib
from scipy.interpolate import griddata
from shapely.geometry import Point

__all__ = ['Waveshuff']

class Waveshuff:
    """
    Sampling implementation of waveshuffling generator
    
    Parameters
    ----------
    x : filename or 1D np.ndarray
        Target brain map
    D : filename or (N,N) np.ndarray or np.memmap
        Pairwise distance matrix between elements of 'x'. Each row of 'D' should
        be sorted. Indices used to sort each row are passed to the 'index' argument.
        index : filename or (N,N) np.ndarray or np.memmap
    vx : filename 
        Path to flatmap gifti file
    ns : int, default 500
        Take a subsample of 'ns' rows from 'D' when fitting variograms
    deltas : np.ndarray or List[float], default [0.3, 0.5, 0.7, 0.9]
        Proportions of neighbors to include for smoothing, in (0, 1)
    pv : int, default 70
        Percentile of the pairwise distance distribution (in 'D') at which to truncate
        during variogram fitting
    nh : int, default 25
        Number of uniformly space distances at which to compute variogram
    knn : int, default 1000
        Number of nearest regions to keep in the neighborhood of each region
    b : float or None, default None
        Gaussian kernel bandwidth for variogram smoothing. If None,
        three times the distance interval spacing is used.
    scales : int or (N,) np.ndarray
        Scale on which to perform wavelet transform (default 1:8)
    wv : string, default 'db8'
        Wavelet transform, accepts ...
    B : float, default 0
        Size of boundary to resample separately
    seed : None or int or np.random.RandomState instance (default 1)
        Specify the seed for random number generation (or random state instance)
    n_jobs : int (default 1)
        Number of jobs to use for parallelizing creation of surrogate maps
    waveamp : int or None, default 1
        Do wavelet amplitude-adjusted step
    """

    def __init__(self, x, D, index, vx=None, ns=500, pv=70, nh=25, knn=1000, b=None,
                 deltas=np.arange(0.3, 1., 0.2), scales=np.arange(8), wv='db8',
                 B=0, seed=1, n_jobs=1, ampadj=True, ch=None):

        self._rs = check_random_state(seed)
        self._n_jobs = n_jobs

        self.x = x
        n = self._x.size
        self.nmap = int(n)
        self.knn = knn
        self.D = D
        self.index = index
        self.vx = vx
        self.nh = int(nh)
        self.deltas = deltas
        self.ns = int(ns)
        self.b = b
        self.pv = pv
        self._ikn = np.arange(self._nmap)[:, None]
        self.ch = ch
        
        self._dmax = np.percentile(self._D, self._pv)
        self.h = np.linspace(self._D.min(), self._dmax, self._nh)
        self.scales = scales
        self.wv = wv
        self.B = B
        self.ampadj = True
        
        if vx:
            self.vertices = load_mesh_vertices(vx)
        else:
            self.vertices = load_mesh_vertices('/Volumes/Scratch/Nik_data/MNINonLinear/fsaverage_LR32k/1001_01_MR.L.flat.32k_fs_LR.surf.gii')

        if not self._b:
            self.b = 3 * (self.h[1] - self.h[0])
        
        #make grid
        self.s, self.gridX, self.gridY = grid(self._x, self.vertices, tol=40, spacing=0.4, ch=self.ch)

    def __call__(self, n=1):
        """
        Generate new surrogate map(s).
        
        Parameters
        ----------
        n : int, default 1
            Number of waveshuffled surrogate maps to generate
        
        Returns
        -------
        (n,N) np.ndarray
            Generated map(s) with matched spatial autocorrelation using wavelets
        
        """
        
        rs = np.arange(n)
        surrs = np.row_stack(
            Parallel(self._n_jobs, prefer='threads')(
                delayed(self._call_method)(rs=i) for i in rs
            )
        )
        return np.asarray(surrs.squeeze())

    def _call_method(self, rs=None):
        """
        Subfunction used by .__call__() for parallelization purposes 
        """        
        #generate waveshuffled map
        x_perm = self.waveamp2()
        
        #de-mean
        x_perm = x_perm - np.nanmean(x_perm)
        
        if self._ismasked:
            return np.ma.masked_array(
                data=x_perm, mask=np.isnan(x_perm)).squeeze()
        
        return x_perm.squeeze()

    def compute_variogram(self, x, idx):
        """
        Compute variogram of 'x' using pairs of regions indexed by 'idx'.
        
        Parameters
        ----------
        x : (N,) np.ndarray
            Brain map
        idx : (ns,) np.ndarray[int]
            Indices of randomly sampled brain regions
        
        Returns
        -------
        v : (ns,ns) np.ndarray
            Variogram y-coordinates, i.e. 0.5 * (x_i - x_j) ^ 2, for i,j in idx
        
        """
        diff_ij = x[idx][:, None] - x[self._index[idx, :]]
        return 0.5 * np.square(diff_ij)

    def smooth_variogram(self, u, v, return_h=False):
        """
        Smooth a variogram.
        
        Parameters
        ----------
        u : (N,) np.ndarray
            Pairwise distances, i.e., variogram x-coordinates
        v : (N,) np.ndarray
            Squared differences, i.e., variogram y-coordinates
        return_h : bool, default False
            Return distances at which smooth variogram is computed
        
        Returns
        -------
        (nh,) np.ndarray
            Smoothed variogram samples
        (nh,) np.ndarray
            Distances at which smoothed variogram was computed (returned if
                                                                'return_h' is True)
        
        """
        if len(u) != len(v):
            raise ValueError("u and v must have the same number of elements")
        
        # Subtract each element of h from each pairwise distance `u`.
        # Each row corresponds to a unique h.
        du = np.abs(u - self._h[:, None])
        w = np.exp(-np.square(2.68 * du / self._b) / 2)
        denom = np.nansum(w, axis=1)
        wv = w * v[None, :]
        num = np.nansum(wv, axis=1)
        output = num / denom
        if not return_h:
            return output
        return output, self._h

    def waveamp2(self):
        """
        Returns
        -------
        surr : (N,) np.ndarray
            Surrogate brain map
        
        """
        s = self.s
        gridX = self.gridX
        gridY = self.gridY 
        n = self._scales
        wv = self.wv
        #B = self._B
        #S = self.seed
        
        rr, col = s.shape
        #dim = 1
        N = max(n)+1
        
        if len(s.shape) > 2:
            raise ValueError("Grid must be 2-D (i.e. flatmap)")
        
        #define random shuffling
        #st = np.floor(np.random.choice(1, len(n))*10^6)
        # fnd = np.where(np.ones((rr, col)))
        
        # #find where edges of data are
        # fnd = np.intersect1d(fnd, np.where(s))
        
        #perform wavelet decomposition
        C = pywt.wavedec2(s, wv, mode='zero', level=N)
        CC = [C[0]]
        
        #cnt = 0
        
        for i in np.arange(1, N+1):
            ch, cv, cd = C[i]
            ch_shuff = shuffle_matrix(ch)
            cv_shuff = shuffle_matrix(cv)
            cd_shuff = shuffle_matrix(cd)
        
            cc = (ch_shuff, cv_shuff, cd_shuff)
        
            CC.append(cc)
            
        #recover waveshuffled grid
        ff = pywt.waverec2(CC, wv, mode='zero')
        f = np.zeros((s.shape))
        #f[np.where(s)] = ff[np.where(s)]
            
        if self.ampadj is True:
            #perform amplitude adjustment (force values of ff to equal values of s)
            f_vec = ff.flatten()
            s_vec = s.flatten()
            
            z = np.zeros((len(f_vec), 3))
            s_st = np.sort(s_vec)
            z[:, 0] = f_vec
            z[:, 1] = np.arange(len(f_vec))
            z = z[np.argsort(z[:, 0])]
            z[:, 2] = s_st
            z = z[np.argsort(z[:, 1])]
            f_vec = z[:, 2]
            
            #f[np.where(s)] = f_vec
            
            f = f_vec.reshape(f.shape)
        
        #transform back to mesh
        meshsurr = mesh(f, self.vertices, gridX, gridY)
        
        return meshsurr
    
    def sample(self):
        """
        Randomly sample (without replacement) brain areas for variogram
        computation.

        Returns
        -------
        (self.ns,) np.ndarray
            Indices of randomly sampled areas

        """
        return self._rs.choice(
            a=self._nmap, size=self._ns, replace=False).astype(np.int32)

    @property
    def x(self):
        """ (N,) np.ndarray : brain map scalars """
        if self._ismasked:
            return np.ma.copy(self._x)
        return np.copy(self._x)
    
    @x.setter
    def x(self, x):
        self._ismasked = False
        x_ = dataio(x)
        check_map(x=x_)
        mask = np.isnan(x_)
        if mask.any():
            self._ismasked = True
            brain_map = np.ma.masked_array(data=x_, mask=mask)
        else:
            brain_map = x_
        self._x = brain_map
    
    @property
    def D(self):
        """ (N,N) np.memmap : Pairwise distance matrix """
        return np.copy(self._D)
    
    @D.setter
    def D(self, x):
        x_ = dataio(x)
        n = self._x.size
        if x_.shape[0] != n:
            raise ValueError(
                "D size along axis=0 must equal brain map size")
        self._D = x_[:, 1:self._knn + 1]  # prevent self-coupling
    
    @property
    def index(self):
        """ (N,N) np.memmap : indexes used to sort each row of dist. matrix """
        return np.copy(self._index)
    
    @index.setter
    def index(self, x):
        x_ = dataio(x)
        n = self._x.size
        if x_.shape[0] != n:
            raise ValueError(
                "index size along axis=0 must equal brain map size")
        self._index = x_[:, 1:self._knn+1].astype(np.int32)
    
    @property
    def nmap(self):
        """ int : length of brain map """
        return self._nmap
    
    @nmap.setter
    def nmap(self, x):
        self._nmap = int(x)
    
    @property
    def pv(self):
        """ int : percentile of pairwise distances at which to truncate """
        return self._pv
    
    @pv.setter
    def pv(self, x):
        pv = check_pv(x)
        self._pv = pv
    
    @property
    def deltas(self):
        """ np.ndarray or List[float] : proportions of nearest neighbors """
        return self._deltas
    
    @deltas.setter
    def deltas(self, x):
        check_deltas(deltas=x)
        self._deltas = x
    
    @property
    def nh(self):
        """ int : number of variogram distance intervals """
        return self._nh
    
    @nh.setter
    def nh(self, x):
        self._nh = x
    
    @property
    def knn(self):
        """ int : number of nearest neighbors included in distance matrix """
        return self._knn
    
    @knn.setter
    def knn(self, x):
        if x > self._nmap:
            raise ValueError('knn must be less than len(X)')
        self._knn = int(x)
        
    @property
    def ns(self):
        """ int : number of randomly sampled regions used to construct map """
        return self._ns

    @ns.setter
    def ns(self, x):
        self._ns = int(x)
    
    @property
    def b(self):
        """ numeric : Gaussian kernel bandwidth """
        return self._b
    
    @b.setter
    def b(self, x):
        self._b = x
    
    @property
    def h(self):
        """ np.ndarray : distances at which variogram is evaluated """
        return self._h
    
    @h.setter
    def h(self, x):
        self._h = x
    
    @property
    def scales(self):
        return self._scales
    
    @scales.setter
    def scales(self, x):
        self._scales = x
    
    @property
    def wv(self):
        return self._wv
    
    @wv.setter
    def wv(self, x):
        self._wv = x
    
    @property
    def B(self):
        return self._B
    
    @B.setter
    def B(self, x):
        self._B = x

        


'''
    setup.py file for neuroshape module
'''

from setuptools import setup, Extension
import numpy

# define the extension module
eta = Extension('neuroshape.eta', sources=['src/eta_squared.c'],
                          include_dirs=[numpy.get_include()])

glmfit = Extension('neuroshape._stats', sources=['src/glmfit.c'],
                   include_dirs=[numpy.get_include()])

euler = Extension('neuroshape.euler', sources=['src/euler_threshold.c'],
                  include_dirs=[numpy.get_include()])

# run the setup
setup(name='neuroshape',
      version='0.0.4.1',
      description="For computing connectopic and geometric Laplacian eigenmodes and performing null hypothesis testing. As implementation is ongoing, this description is subject to rapid change.",
      author='Nikitas C. Koussis, Systems Neuroscience Group Newcastle',
      author_email='nikitas.koussis@gmail.com',
      url='https://github.com/nikitas-k/neuroshape-dev',
      packages=['neuroshape', 'neuroshape.nulls', 'neuroshape.utils'],
      ext_modules=[eta, glmfit, euler])
